"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class SwitchPower1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:SwitchPower:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'SwitchPower1'

    SERVICE_DEFAULT_VARIABLES = {
        "Status": { "data_type": "boolean", "default": "0", "allowed_list": None},
        "Target": { "data_type": "boolean", "default": "0", "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {}

    def action_GetStatus(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetStatus action.

            :returns: "ResultStatus"
        """
        arguments = { }

        out_params = self.call_action("GetStatus", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ResultStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTarget(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTarget action.

            :returns: "RetTargetValue"
        """
        arguments = { }

        out_params = self.call_action("GetTarget", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RetTargetValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetTarget(self, newTargetValue, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetTarget action.
        """
        arguments = {
            "newTargetValue": newTargetValue,
        }

        self.call_action("SetTarget", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional

import os

from akit.exceptions import AKitConfigurationError
from akit.interop.credentials.basecredential import BaseCredential

class SshCredential(BaseCredential):
    """
        The :class:`SshCredential` is a container object for SSH credentials passed in the landscape
        configuration file.

        .. code:: yaml
            "identifier": "straw-node"
            "scheme": "ssh"
            "username": "ubuntu"
            "password": "@@&_@@&_LetMeComeIn"
            "keyfile": "~/.ssh/id_blah_rsa"
            "keypasswd": "@@&_@@&_LetMeComeIn"
            "allow_agent": False
    """

    def __init__(self, *, identifier: str, category: str, role: Optional[str] = "priv", username: str = "",
                 password: Optional[str] = None, keyfile: Optional[str] = None, keypasswd: Optional[str] = None,
                 allow_agent: bool = False, primitive: bool=False):
        """
            :param identifier: The identifier that is used to reference this credential.  (required)
            :param category: The category of credential.
            :param role: An optional parameter that identifies the role that the credential is assigned to.
            :param username: The username associated with this credential.
            :param password: The password associated with this credential.  A password is not required if a
                             keyfile parameter is provided or if 'allow_agent' is passed as 'True'.
            :param keyfile: The private key file to use for authentication with this credential.  A keyfile
                            is not required if a password was passed or if 'allow_agent' is passed as 'True'.
            :param keypasswd: The password to decrypt the keyfile if required by the keyfile.
            :param allow_agent: Indicates if the SSH Agent can be used to authenticate connections.
            :param primitive: When True, simulate file transfers and directory services with ssh commands.
        """
        super().__init__(identifier=identifier, category=category, role=role)

        if category != "ssh":
            raise ValueError("The SshCredential should only be given credentials of category 'ssh'.")
        if len(username) == 0:
            raise ValueError("The SshCredential constructor requires a 'username' parameter be provided.")
        if password is None and keyfile is None and not allow_agent:
            raise ValueError("The SshCredential constructor requires one of: 'password is not None', 'keyfile is not None', 'allow_agent == True'.")

        self._username = username
        self._password = password
        self._keyfile = keyfile
        self._keypasswd = keypasswd
        self._allow_agent = allow_agent
        self._primitive = primitive
        return

    @property
    def allow_agent(self):
        return self._allow_agent

    @property
    def keyfile(self):
        return self._keyfile

    @property
    def keypasswd(self):
        return self._keypasswd

    @property
    def password(self):
        return self._password

    @property
    def primitive(self):
        return self._primitive

    @property
    def username(self):
        return self._username

    @classmethod
    def validate(cls, cred_info):

        errmsg_lines = []

        allow_agent = False
        if "allow_agent" in cred_info:
            allow_agent = cred_info["allow_agent"]

        if "identifier" not in cred_info:
            errmsg_lines.append("    * missing 'identifier' parameter")

        if "username" not in cred_info:
            errmsg_lines.append("    * missing 'username' parameter")

        if "password" not in cred_info and "keyfile" not in cred_info and not allow_agent:
                errmsg_lines.append("    * missing 'password' or 'keyfile' when allow_agent is 'False'")

        if "keyfile" in cred_info:
            keyfile = os.path.abspath(os.path.expandvars(os.path.expanduser(cred_info["keyfile"])))
            if not os.path.exists(keyfile):
                errmsg_lines.append("    * specified 'keyfile={}' not found.".format(keyfile))

        if len(errmsg_lines) > 0:
            identifier = "????"
            if "identifier" in cred_info:
                identifier = cred_info["identifier"]

            errmsg = "Errors found while validating the '{}' SSH credential:".format(identifier)
            errmsg_lines.insert(0, errmsg)
            errmsg = os.linesep.join(errmsg_lines)

            raise AKitConfigurationError(errmsg) from None

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Any, List, Protocol

from akit.exceptions import AKitNotImplementedError

class IIncludeFilter(Protocol):
    """
        The IIncludeFilter interface is used to provide a common interface for performing an
        include filter on objects.
    """

    def should_include(self, check_object: Any) -> bool:
        """
            Determines if a device matches an include criteria internalized in the filter object.

            :param check_object: The object to check for a match with the exclude criteria.
        """
    
    def filter(self, input_list: List[Any]) -> List[Any]:

        output_list = []

        for nxt_object in input_list:
            if self.should_include(nxt_object):
                output_list.append(nxt_object)

        return output_list



from typing import (
    Any, Callable, Dict, List, get_args as typing_get_args
)

import collections

from akit.exceptions import AKitSemanticError
from akit.coupling.integrationcoupling import IntegrationCoupling
from akit.testing.testplus.scopecoupling import ScopeCoupling
from akit.testing.testplus.registration.parameterorigin import ParameterOrigin

from akit.testing.testplus.testref import TestRef

RESOURCE_KEY_FORMAT = "{}@{}:{}"

RESERVED_PARAMETER_NAMES = ["constraints"]

class ResourceScope:

    def __init__(self, name=None, package=None, is_test_scope=False):
        self._name = name
        self._package = package
        self._children = {}
        self._parameter_originations = collections.OrderedDict()
        self._is_test_scope = is_test_scope
        self._is_relevant = False
        return

    @property
    def children(self) -> Dict[str, "ResourceScope"]:
        return self._children

    @property
    def is_relevant(self) -> bool:
        return self._is_relevant

    @property
    def name(self) -> str:
        return self._name

    @property
    def package(self) -> str:
        return self._package

    @property
    def parameter_originations(self) -> Dict[str, ParameterOrigin]:
        return self._parameter_originations

    def add_descendent_parameter_origination(self, assigned_scope: str, parameter_origin: ParameterOrigin):
        """
            Adds a descendent parameter origin object into the parameter origin table of the
            assigned scope path provided.

            :param assigned_scope: The full scope name of the scope to assign the parameter
                                   origin to.
            :param parameter_origin: The parameter origin object that is being inserted into
                                     the parameter origin table of the assigned scope.
        """

        if self._package is not None:
            err_msg = "The 'add_descendent' API can only be called on the root package."
            raise AKitSemanticError(err_msg) from None

        if self._package is None and assigned_scope == "<session>":
            if parameter_origin is not None:
                identifier = parameter_origin.identifier
                self._parameter_originations[identifier] =  parameter_origin
                self._parameter_originations.move_to_end(identifier, last=True)
        else:
            is_test_scope = False
            if assigned_scope.find("#") > -1:
                is_test_scope = True
                assigned_scope = assigned_scope.replace("#", ".")
            to_walk_list = assigned_scope.split(".")
            path_stack = []

            self._add_descendent_parameter_origination(parameter_origin, to_walk_list, path_stack, is_test_scope)

        return

    def collect_integrations_and_scopes_for_test(self, integration_table: Dict[Callable, IntegrationCoupling],
        scope_table: Dict[Callable, ScopeCoupling]):
        """
            Walk the scope tree path for the test reference provide and collect all the integration and
            scope couplings that are referenced along the path.
        """

        for param_orig in self._parameter_originations.values():

            source_function = param_orig.source_function
            param_resource_type = param_orig.source_resource_type

            if len(typing_get_args(param_resource_type)) > 0:
                for type_arg in typing_get_args(param_resource_type):
                    if issubclass(type_arg, IntegrationCoupling):
                        # There should never be more than one fixture with the same well-known or declared name in
                        # the same collection of tests.
                        integration_table[source_function] = type_arg
                        break
                    elif issubclass(type_arg, ScopeCoupling):
                        scope_table[source_function] = type_arg
                        break
            else:
                if issubclass(param_resource_type, IntegrationCoupling):
                    # There should never be more than one fixture with the same well-known or declared name in
                    # the same collection of tests.
                    integration_table[source_function] = param_resource_type
                elif issubclass(param_resource_type, ScopeCoupling):
                    scope_table[source_function] = param_resource_type

        for child_scope in self._children.values():

            # Call each child and collect descendant integrations and scopes.
            child_scope.collect_integrations_and_scopes_for_test(integration_table, scope_table)

        return

    def ensure_parameter_scopes_for_test(self, test_ref: TestRef):
        """
            Walks through the scope tree using the full scope name of a test reference
            object and esures there is a scope node for each component of the test reference
            path name.

            :param test_ref: The :class:`TestRef` object that ancestrial scope nodes should
                             be created for.
        """

        test_scope_name = test_ref.scope_name

        module_name, test_name = test_scope_name.split("#")

        test_scope_parts = module_name.split(".")
        test_scope_parts.append(test_name)

        path_stack = []

        self._is_relevant = True
        self._ensure_parameter_scopes_for_test(test_ref, test_scope_parts, path_stack)

        return

    def has_descendent_parameter(self, scope_name: str, identifier: str) -> bool:
        """
            Method used to findout if the specified scope has a parameter or not
            available 

            :param scope_name: The full path of the scope name to lookup.
            :param identifier: The name of the identifier to search for.
        """

        rtnval = False

        if self._package is None and scope_name == "<session>":
            if identifier in self._parameter_originations:
                rtnval = True
        else:
            scope_name = scope_name.replace("#", ".")
            to_walk_list = scope_name.split(".")

            rtnval = self._has_descendent_parameter(to_walk_list, identifier)

        return rtnval

    def lookup_scope(self, scope_name: str) -> "ResourceScope":
        """
            Method used to parse a full scope name into components and then lookup
            the specified scope in the scope tree.

            :param scope_name: The full path of the scope name to lookup.
        """

        scope_found = None
        if self._package is None and scope_name == "<session>":
            scope_found = self
        else:
            scope_name = scope_name.replace("#", ".")
            to_walk_list = scope_name.split(".")

            if len(to_walk_list) > 0:
                scope_found = self._lookup_scope(to_walk_list)

        return scope_found

    def prune_unreference_scopes(self):
        """
            Walk through all of the child nodes and prune any not relevant scope nodes
            and any of thier descendants.
        """

        temp_child_key_list = [csk for csk in self._children.keys()]
        while len(temp_child_key_list) > 0:
            nxt_child_key = temp_child_key_list.pop()
            nxt_child = self._children[nxt_child_key]

            # Have this child prune its children before we prune it
            # from this nodes children table
            if len(nxt_child._children) > 0:
                nxt_child.prune_unreference_scopes()

            # If we had this child prune its children, then we can
            # prune this child.
            if not nxt_child._is_relevant:
                del self._children[nxt_child_key]

        return

    def rename_resource_origins_from_main(self, new_origin: str):
        """
        """

        if "__main__" in self._children:
            last_dot = new_origin.rfind(".")

            cobj = self._children["__main__"]
            del self._children["__main__"]

            self._children[new_origin] = cobj

            # All of the children of cobj will be tests, go through all of
            # the child objects children and fix the package name
            for testscope in cobj._children.values():
                test_scope_origin = "{}#{}".format(new_origin, testscope._name)
                for _, poval in testscope.parameter_originations.items():
                    poval._originating_scope = test_scope_origin
                    self.add_descendent_parameter_origination(test_scope_origin, poval)

        return

    def resolve_parameter_originations_for_test(self, test_ref: TestRef, missing_params: List[Any]):
        """
            Walk through all the descendant nodes of this node utilizing the test reference
            full scope name for a path and validate the resolution of all parameters that are
            being utilized by any parameter source functions or any of thier sources.

            :param test_ref: The :class:`TestRef` object whos scope path is used to validate a
                             single scope path from the set of all parameter scope paths.
            :param missing_params: The list of parameters that are found to be missing when
                                   searching the descent of the scope path.
        """

        test_scope_name = test_ref.scope_name

        module_name, test_name = test_scope_name.split("#")

        test_scope_parts = module_name.split(".")
        test_scope_parts.append(test_name)

        param_table = {}
        for pname, pinfo in self._parameter_originations.items():
            param_table[pname] = pinfo

        self._resolve_parameter_originations_for_test_descend(test_ref, test_scope_parts, param_table, missing_params)

        return

    def _add_descendent_parameter_origination(self, parameter_origin: ParameterOrigin, to_walk_list: List[str], path_stack: List[str], is_test_scope):

        curr_leaf = to_walk_list.pop(0)
        curr_scope = None

        if curr_leaf in self._children:
            curr_scope = self._children[curr_leaf]
        else:
            curr_package = curr_leaf
            if len(path_stack) > 0:
                curr_package = "{}.{}".format(".".join(path_stack), curr_package)

            set_test_scope = False
            if len(to_walk_list) == 0:
                set_test_scope = is_test_scope

            curr_scope = ResourceScope(curr_leaf, curr_package, set_test_scope)
            self._children[curr_leaf] = curr_scope

        if len(to_walk_list) > 0:
            path_stack.append(curr_leaf)
            curr_scope._add_descendent_parameter_origination(parameter_origin, to_walk_list, path_stack, is_test_scope)
            path_stack.pop()
        else:
            if parameter_origin is not None:
                identifier = parameter_origin.identifier
                curr_scope._parameter_originations[identifier] = parameter_origin
                curr_scope._parameter_originations.move_to_end(identifier, last=True)

        return

    def _ensure_parameter_scopes_for_test(self, test_ref: TestRef, test_scope_parts: List[str], path_stack: List[str]):

        if len(test_scope_parts) > 0:
            curr_leaf = test_scope_parts.pop(0)

            if curr_leaf not in self._children:
                curr_package = curr_leaf
                if len(path_stack) > 0:
                    curr_package = "{}.{}".format(".".join(path_stack), curr_package)
                
                set_test_scope = False
                if len(test_scope_parts) == 0:
                    set_test_scope = True
                curr_scope = ResourceScope(curr_leaf, curr_package, set_test_scope)
                curr_scope._is_relevant = True
                self._children[curr_leaf] = curr_scope

                path_stack.append(curr_leaf)
                curr_scope = self._children[curr_leaf]
                curr_scope._ensure_parameter_scopes_for_test(test_ref, test_scope_parts, path_stack)

            else:
                path_stack.append(curr_leaf)
                curr_scope = self._children[curr_leaf]
                curr_scope._is_relevant = True
                curr_scope._ensure_parameter_scopes_for_test(test_ref, test_scope_parts, path_stack)

        return

    def _has_descendent_parameter(self, to_walk_list: List[str], identifier: str):

        rtnval = False

        if len(to_walk_list) == 0:
            if identifier in self._parameter_originations:
                rtnval = True
        else:
            child_name = to_walk_list[0]
            if child_name in self._children:
                scope = self._children[child_name]
                desc_to_walk_list = to_walk_list[1:]
                rtnval = scope._has_descendent_parameter(desc_to_walk_list, identifier)

        return rtnval

    def _lookup_scope(self, to_walk_list: List[str]) -> "ResourceScope":
        """
            The private recursive method for lookup_scope that is used to find a scope by path
            in the scope tree.

            :param to_walk_list: A list of scope names to walk down to find the desired scope.
        """

        scope_found = None

        child_name = to_walk_list.pop(0)
        if child_name in self._children:
            child_scope = self._children[child_name]
            if len(to_walk_list) > 0:
                scope_found = child_scope._lookup_scope(to_walk_list)
            else:
                scope_found = child_scope

        return scope_found

    def _resolve_parameter_originations_for_test_descend(self, test_ref: str, test_scope_parts: List[str],
        param_table: dict, missing_params: List[Any]):

        if len(test_scope_parts) > 0:
            nxt_scope_name = test_scope_parts.pop(0)
            if nxt_scope_name in self._children:
                nxt_scope = self._children[nxt_scope_name]

                nxt_param_table = param_table.copy()

                # Go through all the parameter originations registered in this scope an add them
                # to the nxt_param_table dictionary to pass down to descendant scopes
                origination_items = [(spn, spo) for spn, spo in nxt_scope.parameter_originations.items()]
                for scope_param_name, scope_param_origin in origination_items:
                    
                    if isinstance(scope_param_origin, ParameterOrigin):
                        # Do a lateral check to make sure the originating methods of any declared parameters
                        # already have parameters declared in the current in scope parameters
                        self._resolve_parameter_originations_for_test_lateral(test_ref, scope_param_origin,
                            test_scope_parts, nxt_param_table, missing_params)

                        # Add the new parameter origination to the parameter lookup table for this scope as
                        # it came first and can feed any following parameter source.
                        nxt_param_table[scope_param_name] = scope_param_origin
                    else:
                        print("{}:{}".format(nxt_scope.package, scope_param_name))

                nxt_scope._resolve_parameter_originations_for_test_descend(test_ref, test_scope_parts, nxt_param_table, missing_params)
            else:
                print("child not found '{}'".format(nxt_scope_name))
        return

    def _resolve_parameter_originations_for_test_lateral(self, test_ref: str, lateral_param_origin: ParameterOrigin,
        test_scope_parts: List[str], scope_param_table: dict, missing_params: List[Any]):

        source_params = lateral_param_origin.source_signature.parameters
        if len(source_params) > 0:
            for sparam_name, sparam_origin in source_params.items():
                if sparam_name not in scope_param_table and sparam_name not in RESERVED_PARAMETER_NAMES:
                    missing_params.append([sparam_name, sparam_origin])
        return




"""
.. module:: shellscript
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A module that provides the ShellScript task class which implements
               the execution of shell based tasks in a workpacket.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional

import os
import stat
import subprocess

from akit.paths import get_temporary_file
from akit.xformatting import indent_script

from akit.workflow.tasks.taskbase import TaskBase

class ShellScript(TaskBase):

    def __init__(self, ordinal: str, label: str, task_info: dict, logger):
        super(ShellScript, self).__init__(ordinal, label, task_info, logger)
        self._script = task_info["script"]
        return

    @property
    def script(self) -> str:
        return self._script

    def execute(self, parameters: Optional[dict]=None, topology: Optional[dict]=None, **kwargs) -> int:

        self._logger.info("STEP: %s - %s" % (self._label, self._ordinal))

        script_content = os.linesep.join(self._script)

        tempfile = get_temporary_file(prefix="step-%d" % self._ordinal, suffix=".sh")

        self._logger.info("Running Script: %s" % tempfile)

        with open(tempfile, 'w') as tf:
            tf.write(script_content)

        os.chmod(tempfile, stat.S_IEXEC)

        proc = subprocess.Popen([tempfile], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = proc.communicate()
        exit_code = proc.wait()

        log_msg_script = [
            "RESULT CODE: %d" % exit_code,
            "STDOUT:",
            indent_script(stdout.read(), level=1),
            "STDERR:",
            indent_script(stderr.read(), level=1),
        ]

        log_msg = os.linesep.join(log_msg_script)
        if exit_code == 0:
            self._logger.info(log_msg)
        else:
            self._logger.error(log_msg)

        return exit_code


"""
.. module:: gitcheckout
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A module that provides the ShellScript task class which implements
               the execution of shell based tasks in a workpacket.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional

from akit.workflow.tasks.taskbase import TaskBase

class GitCheckout(TaskBase):

    def __init__(self, ordinal: str, label: str, task_info: dict, logger):
        super(GitCheckout, self).__init__(ordinal, label, task_info, logger)
        return

    def execute(self, parameters: Optional[dict]=None, topology: Optional[dict]=None, **kwargs) -> int:

        self._logger.info("STEP: %s - %s" % (self._label, self._ordinal))

        exit_code = 0
        return exit_code


"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class ContentDirectory1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'ContentDirectory1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:ContentDirectory:1'

    SERVICE_DEFAULT_VARIABLES = {
        "SearchCapabilities": { "data_type": "string", "default": None, "allowed_list": None},
        "SortCapabilities": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "Browseable": { "data_type": "boolean", "default": None, "allowed_list": None},
        "ContainerUpdateIDs": { "data_type": "string", "default": None, "allowed_list": None},
        "FavoritePresetsUpdateID": { "data_type": "string", "default": None, "allowed_list": None},
        "FavoritesUpdateID": { "data_type": "string", "default": None, "allowed_list": None},
        "RadioFavoritesUpdateID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "RadioLocationUpdateID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "RecentlyPlayedUpdateID": { "data_type": "string", "default": None, "allowed_list": None},
        "SavedQueuesUpdateID": { "data_type": "string", "default": None, "allowed_list": None},
        "ShareIndexInProgress": { "data_type": "boolean", "default": None, "allowed_list": None},
        "ShareIndexLastError": { "data_type": "string", "default": None, "allowed_list": None},
        "ShareListUpdateID": { "data_type": "string", "default": None, "allowed_list": None},
        "SystemUpdateID": { "data_type": "ui4", "default": None, "allowed_list": None},
        "UserRadioUpdateID": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_Browse(self, ObjectID, BrowseFlag, Filter, StartingIndex, RequestedCount, SortCriteria, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the Browse action.

            :returns: "Result", "NumberReturned", "TotalMatches", "UpdateID"
        """
        arguments = {
            "ObjectID": ObjectID,
            "BrowseFlag": BrowseFlag,
            "Filter": Filter,
            "StartingIndex": StartingIndex,
            "RequestedCount": RequestedCount,
            "SortCriteria": SortCriteria,
        }

        out_params = self.call_action("Browse", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Result", "NumberReturned", "TotalMatches", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_CreateObject(self, ContainerID, Elements, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CreateObject action.

            :returns: "ObjectID", "Result"
        """
        arguments = {
            "ContainerID": ContainerID,
            "Elements": Elements,
        }

        out_params = self.call_action("CreateObject", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ObjectID", "Result",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DestroyObject(self, ObjectID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DestroyObject action.
        """
        arguments = {
            "ObjectID": ObjectID,
        }

        self.call_action("DestroyObject", arguments=arguments, aspects=aspects)

        return

    def action_FindPrefix(self, ObjectID, Prefix, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the FindPrefix action.

            :returns: "StartingIndex", "UpdateID"
        """
        arguments = {
            "ObjectID": ObjectID,
            "Prefix": Prefix,
        }

        out_params = self.call_action("FindPrefix", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("StartingIndex", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAlbumArtistDisplayOption(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAlbumArtistDisplayOption action.

            :returns: "AlbumArtistDisplayOption"
        """
        arguments = { }

        out_params = self.call_action("GetAlbumArtistDisplayOption", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("AlbumArtistDisplayOption",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetAllPrefixLocations(self, ObjectID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetAllPrefixLocations action.

            :returns: "TotalPrefixes", "PrefixAndIndexCSV", "UpdateID"
        """
        arguments = {
            "ObjectID": ObjectID,
        }

        out_params = self.call_action("GetAllPrefixLocations", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("TotalPrefixes", "PrefixAndIndexCSV", "UpdateID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetBrowseable(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBrowseable action.

            :returns: "IsBrowseable"
        """
        arguments = { }

        out_params = self.call_action("GetBrowseable", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("IsBrowseable",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetLastIndexChange(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetLastIndexChange action.

            :returns: "LastIndexChange"
        """
        arguments = { }

        out_params = self.call_action("GetLastIndexChange", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("LastIndexChange",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSearchCapabilities(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSearchCapabilities action.

            :returns: "SearchCaps"
        """
        arguments = { }

        out_params = self.call_action("GetSearchCapabilities", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SearchCaps",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetShareIndexInProgress(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetShareIndexInProgress action.

            :returns: "IsIndexing"
        """
        arguments = { }

        out_params = self.call_action("GetShareIndexInProgress", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("IsIndexing",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSortCapabilities(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSortCapabilities action.

            :returns: "SortCaps"
        """
        arguments = { }

        out_params = self.call_action("GetSortCapabilities", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("SortCaps",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSystemUpdateID(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSystemUpdateID action.

            :returns: "Id"
        """
        arguments = { }

        out_params = self.call_action("GetSystemUpdateID", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Id",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RefreshShareIndex(self, AlbumArtistDisplayOption, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RefreshShareIndex action.
        """
        arguments = {
            "AlbumArtistDisplayOption": AlbumArtistDisplayOption,
        }

        self.call_action("RefreshShareIndex", arguments=arguments, aspects=aspects)

        return

    def action_RequestResort(self, SortOrder, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RequestResort action.
        """
        arguments = {
            "SortOrder": SortOrder,
        }

        self.call_action("RequestResort", arguments=arguments, aspects=aspects)

        return

    def action_SetBrowseable(self, Browseable, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetBrowseable action.
        """
        arguments = {
            "Browseable": Browseable,
        }

        self.call_action("SetBrowseable", arguments=arguments, aspects=aspects)

        return

    def action_UpdateObject(self, ObjectID, CurrentTagValue, NewTagValue, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UpdateObject action.
        """
        arguments = {
            "ObjectID": ObjectID,
            "CurrentTagValue": CurrentTagValue,
            "NewTagValue": NewTagValue,
        }

        self.call_action("UpdateObject", arguments=arguments, aspects=aspects)

        return




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import argparse
import os
import sys

import click

from akit.environment.variables import LOG_LEVEL_NAMES

@click.group("jobs", help="Contains commands for working with test jobs.")
def group_akit_testing_jobs():
    return

HELP_ROOT = "The root directory to use when scanning for tests."
HELP_JOB = "The identifier of the job to run."
HELP_FILTER = "Add a job filter expression."
HELP_OUTPUT = "The output directory where results and artifacts are collected."
HELP_START = r"A time stamp to associate with the start of the run. Example: 2020-10-17T15:30:11.989120  Bash: date +%Y-%m-%dT%H:%M:%S.%N"
HELP_BRANCH = "The name of the branch to associate with the test run results."
HELP_BUILD = "The name of the build to associate with the test run results."
HELP_FLAVOR = "The name of the flavor to associate with the test run results."
HELP_CONSOLE_LOG_LEVEL = "The logging level for console output."
HELP_FILE_LOG_LEVEL = "The logging level for logfile output."
HELP_DEBUG = "Output debug information to the console."

@click.command("list")
@click.option("--root", default=".", type=str, help=HELP_ROOT)
@click.option("--filter", "-f", required=False, help=HELP_FILTER)
@click.option("--debug", default=False, type=bool, help=HELP_DEBUG)
def command_testing_jobs_list():
    return

@click.command("show")
@click.option("--root", default=".", type=str, help=HELP_ROOT)
@click.option("--job", "-j", required=True, help=HELP_JOB)
@click.option("--debug", default=False, type=bool, help=HELP_DEBUG)
def command_akit_testing_jobs_show():
    return

@click.command("run")
@click.option("--root", default=".", type=str, help=HELP_ROOT)
@click.option("--job", "-j", required=True, help=HELP_JOB)
@click.option("--output", "-o", required=False, help=HELP_OUTPUT)
@click.option("--start", default=None, required=False, help=HELP_START)
@click.option("--branch", default=None, required=False, help=HELP_BRANCH)
@click.option("--build", default=None, required=False, help=HELP_BUILD)
@click.option("--flavor", default=None, required=False, help=HELP_FLAVOR)
@click.option("--console-level", default=None, required=False, type=click.Choice(LOG_LEVEL_NAMES, case_sensitive=False), help=HELP_CONSOLE_LOG_LEVEL)
@click.option("--logfile-level", default=None, required=False, type=click.Choice(LOG_LEVEL_NAMES, case_sensitive=False), help=HELP_FILE_LOG_LEVEL)
def command_akit_testing_jobs_run(root, job, output, start, branch, build, flavor, console_level, logfile_level):

    # pylint: disable=unused-import,import-outside-toplevel

    # We do the imports of the automation framework code inside the action functions because
    # we don't want to startup loggin and the processing of inputs and environment variables
    # until we have entered an action method.  Thats way we know how to setup the environment.

    # IMPORTANT: We need to load the context first because it will trigger the loading
    # of the default user configuration
    from akit.environment.context import Context

    from akit.compat import import_by_name
    from akit.environment.variables import extend_path

    try:
        ctx = Context()
        ctx.insert("/environment/jobtype", 'testrun')

        test_root = os.path.abspath(os.path.expandvars(os.path.expanduser(root)))
        if not os.path.isdir(test_root):
            errmsg = "The specified root folder does not exist. root=%s" % root
            if test_root != root:
                errmsg += " expanded=%s" % test_root
            raise argparse.ArgumentError("--root", errmsg)

        # Make sure we extend PATH to include the test root
        extend_path(test_root)

        # We perform activation a little later in the testrunner.py file so we can
        # handle exceptions in the context of testrunner_main function
        import akit.activation.testrun
        from akit.xlogging.foundations import logging_initialize, getAutomatonKitLogger

        # Initialize logging
        logging_initialize()
        logger = getAutomatonKitLogger()

        from akit.extensionpoints import AKitExtensionPoints
        akep = AKitExtensionPoints()

        # At this point in the code, we either lookup an existing test job or we create a test job
        # from the includes, excludes or test_module
        TestJobType = akep.get_testplus_default_job_type()

        if job is not None:
            job_parts = job.split("@")
            if len(job_parts) != 2:
                errmsg = "A --job parameter must be of the form 'package.module@JobClass'"
                raise click.UsageError(errmsg)

            job_package, job_class = job_parts

            try:
                job_mod = import_by_name(job_package)
            except ImportError as imperr:
                errmsg = "Failure while importing job package %r"  % job_package
                raise argparse.ArgumentError("--job", errmsg) from imperr

            if not hasattr(job_mod, job_class):
                errmsg = "The job package %r does not have a job type %r." % (job_package, job_class)
                raise argparse.ArgumentError("--job", errmsg)

            TestJobType = getattr(job_mod, job_class)

        result_code = 0
        with TestJobType(logger, test_root) as tjob:
            result_code = tjob.execute()

        sys.exit(result_code)

    finally:
        pass

    return

group_akit_testing_jobs.add_command(command_akit_testing_jobs_show)
group_akit_testing_jobs.add_command(command_akit_testing_jobs_run)



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class Dimming1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:Dimming:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'Dimming1'

    SERVICE_DEFAULT_VARIABLES = {
        "LoadLevelTarget": { "data_type": "ui1", "default": "0", "allowed_list": None},
        "OnEffect": { "data_type": "string", "default": "Default", "allowed_list": "['OnEffectLevel', 'LastSetting', 'Default']"},
        "OnEffectLevel": { "data_type": "ui1", "default": "100", "allowed_list": None},
        "RampTime": { "data_type": "ui4", "default": "0", "allowed_list": None},
        "ValidOutputValues": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "IsRamping": { "data_type": "boolean", "default": "0", "allowed_list": None},
        "LoadLevelStatus": { "data_type": "ui1", "default": "0", "allowed_list": None},
        "RampPaused": { "data_type": "boolean", "default": "0", "allowed_list": None},
        "RampRate": { "data_type": "ui1", "default": "0", "allowed_list": None},
        "StepDelta": { "data_type": "ui1", "default": "Manufacturer defined default value", "allowed_list": None},
    }

    def action_GetIsRamping(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetIsRamping action.

            :returns: "retIsRamping"
        """
        arguments = { }

        out_params = self.call_action("GetIsRamping", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("retIsRamping",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetLoadLevelStatus(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetLoadLevelStatus action.

            :returns: "retLoadlevelStatus"
        """
        arguments = { }

        out_params = self.call_action("GetLoadLevelStatus", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("retLoadlevelStatus",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetLoadLevelTarget(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetLoadLevelTarget action.

            :returns: "GetLoadlevelTarget"
        """
        arguments = { }

        out_params = self.call_action("GetLoadLevelTarget", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("GetLoadlevelTarget",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetOnEffectParameters(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetOnEffectParameters action.

            :returns: "retOnEffect", "retOnEffectLevel"
        """
        arguments = { }

        out_params = self.call_action("GetOnEffectParameters", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("retOnEffect", "retOnEffectLevel",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRampPaused(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRampPaused action.

            :returns: "retRampPaused"
        """
        arguments = { }

        out_params = self.call_action("GetRampPaused", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("retRampPaused",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRampRate(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRampRate action.

            :returns: "retRampRate"
        """
        arguments = { }

        out_params = self.call_action("GetRampRate", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("retRampRate",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRampTime(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRampTime action.

            :returns: "retRampTime"
        """
        arguments = { }

        out_params = self.call_action("GetRampTime", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("retRampTime",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetStepDelta(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetStepDelta action.

            :returns: "retStepDelta"
        """
        arguments = { }

        out_params = self.call_action("GetStepDelta", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("retStepDelta",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_PauseRamp(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the PauseRamp action.
        """
        arguments = { }

        self.call_action("PauseRamp", arguments=arguments, aspects=aspects)

        return

    def action_ResumeRamp(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ResumeRamp action.
        """
        arguments = { }

        self.call_action("ResumeRamp", arguments=arguments, aspects=aspects)

        return

    def action_SetLoadLevelTarget(self, newLoadlevelTarget, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetLoadLevelTarget action.
        """
        arguments = {
            "newLoadlevelTarget": newLoadlevelTarget,
        }

        self.call_action("SetLoadLevelTarget", arguments=arguments, aspects=aspects)

        return

    def action_SetOnEffect(self, newOnEffect, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetOnEffect action.
        """
        arguments = {
            "newOnEffect": newOnEffect,
        }

        self.call_action("SetOnEffect", arguments=arguments, aspects=aspects)

        return

    def action_SetOnEffectLevel(self, newOnEffectLevel, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetOnEffectLevel action.
        """
        arguments = {
            "newOnEffectLevel": newOnEffectLevel,
        }

        self.call_action("SetOnEffectLevel", arguments=arguments, aspects=aspects)

        return

    def action_SetRampRate(self, newRampRate, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRampRate action.
        """
        arguments = {
            "newRampRate": newRampRate,
        }

        self.call_action("SetRampRate", arguments=arguments, aspects=aspects)

        return

    def action_SetStepDelta(self, newStepDelta, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetStepDelta action.
        """
        arguments = {
            "newStepDelta": newStepDelta,
        }

        self.call_action("SetStepDelta", arguments=arguments, aspects=aspects)

        return

    def action_StartRampDown(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartRampDown action.
        """
        arguments = { }

        self.call_action("StartRampDown", arguments=arguments, aspects=aspects)

        return

    def action_StartRampToLevel(self, newLoadLevelTarget, newRampTime, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartRampToLevel action.
        """
        arguments = {
            "newLoadLevelTarget": newLoadLevelTarget,
            "newRampTime": newRampTime,
        }

        self.call_action("StartRampToLevel", arguments=arguments, aspects=aspects)

        return

    def action_StartRampUp(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StartRampUp action.
        """
        arguments = { }

        self.call_action("StartRampUp", arguments=arguments, aspects=aspects)

        return

    def action_StepDown(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StepDown action.
        """
        arguments = { }

        self.call_action("StepDown", arguments=arguments, aspects=aspects)

        return

    def action_StepUp(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StepUp action.
        """
        arguments = { }

        self.call_action("StepUp", arguments=arguments, aspects=aspects)

        return

    def action_StopRamp(self, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the StopRamp action.
        """
        arguments = { }

        self.call_action("StopRamp", arguments=arguments, aspects=aspects)

        return



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class HTControl1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'HTControl1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:HTControl:1'

    SERVICE_DEFAULT_VARIABLES = {
        "LEDFeedbackState": { "data_type": "string", "default": None, "allowed_list": "['On', 'Off']"},
        "RemoteConfigured": { "data_type": "boolean", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "IRRepeaterState": { "data_type": "string", "default": None, "allowed_list": "['On', 'Off', 'Disabled']"},
        "TOSLinkConnected": { "data_type": "boolean", "default": None, "allowed_list": None},
    }

    def action_CommitLearnedIRCodes(self, Name, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CommitLearnedIRCodes action.
        """
        arguments = {
            "Name": Name,
        }

        self.call_action("CommitLearnedIRCodes", arguments=arguments, aspects=aspects)

        return

    def action_GetIRRepeaterState(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetIRRepeaterState action.

            :returns: "CurrentIRRepeaterState"
        """
        arguments = { }

        out_params = self.call_action("GetIRRepeaterState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentIRRepeaterState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetLEDFeedbackState(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetLEDFeedbackState action.

            :returns: "LEDFeedbackState"
        """
        arguments = { }

        out_params = self.call_action("GetLEDFeedbackState", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("LEDFeedbackState",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_IdentifyIRRemote(self, Timeout, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the IdentifyIRRemote action.
        """
        arguments = {
            "Timeout": Timeout,
        }

        self.call_action("IdentifyIRRemote", arguments=arguments, aspects=aspects)

        return

    def action_IsRemoteConfigured(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the IsRemoteConfigured action.

            :returns: "RemoteConfigured"
        """
        arguments = { }

        out_params = self.call_action("IsRemoteConfigured", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RemoteConfigured",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_LearnIRCode(self, IRCode, Timeout, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the LearnIRCode action.
        """
        arguments = {
            "IRCode": IRCode,
            "Timeout": Timeout,
        }

        self.call_action("LearnIRCode", arguments=arguments, aspects=aspects)

        return

    def action_SetIRRepeaterState(self, DesiredIRRepeaterState, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetIRRepeaterState action.
        """
        arguments = {
            "DesiredIRRepeaterState": DesiredIRRepeaterState,
        }

        self.call_action("SetIRRepeaterState", arguments=arguments, aspects=aspects)

        return

    def action_SetLEDFeedbackState(self, LEDFeedbackState, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetLEDFeedbackState action.
        """
        arguments = {
            "LEDFeedbackState": LEDFeedbackState,
        }

        self.call_action("SetLEDFeedbackState", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from abc import ABC, abstractmethod

from rich.progress import Progress

from akit.exceptions import AKitNotOverloadedError

class ProcedureContext(ABC):

    def __init__(self, identifier: str):
        self._identifier = identifier
        self._progress_console = None
        return

    @property
    def identifier(self):
        return self._identifier

    @property
    def progress_console(self):
        return self._progress_console

    @abstractmethod
    def attach_to_progress_console(self, progress_console: Progress):
        errmsg = "The ProcedureContext:attach_to_progress_console method must be overloaded by derived types."
        raise AKitNotOverloadedError(errmsg)



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

def broadcast_wake_on_lan_magic_message(brodcast_addr: str, mac_addr: str):
    '[FF FF FF FF FF FF] + [mac] * 16   ( len 102 bytes )'
    return


"""
.. module:: testcollector
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains the `TestCollector` object which is utilized to collection the
               information about the tests, scopes and integration that will be included in an
               automation run.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Dict, List, Sequence, Tuple, Union
from types import ModuleType

import fnmatch
import inspect
import os
import sys
import traceback

from akit.exceptions import AKitSemanticError, AKitUnknownParameterError
from akit.coupling.integrationcoupling import IntegrationCoupling, is_integration_coupling

from akit.testing.expressions import parse_test_include_expression
from akit.testing.utilities import find_included_modules_under_root
from akit.testing.testplus.queries import collect_test_references

from akit.testing.testplus.testgroup import TestGroup
from akit.testing.testplus.testref import TestRef
from akit.testing.testplus.registration.resourceregistry import resource_registry
from akit.testing.testplus.markers import MetaFilter

from akit.xlogging.foundations import getAutomatonKitLogger

logger = getAutomatonKitLogger()

def catagorize_exclusions(excluded: str) -> Tuple[List[str], List[str]]:
    """
        Go through a list of exclusion expressions and divide them up by whether they are
        file based exclusions or an exclusion that targets python types or methods within
        a specific module.

        :param excluded: A list of exclusion expressions.

        :returns: A list of file exclusion expressions and a list of specific exclusions.
    """
    file_exclusions = []
    specific_exclusions = []

    if excluded is not None:
        for ex in excluded:
            if ex.find("@") < -1:
                path_prefix = ex.replace(".", os.sep)
                file_exclusions.append(path_prefix)
            else:
                specific_exclusions.append(ex)

    return file_exclusions, specific_exclusions

class TestCollector:
    """
        The :class:`TestCollector` object utilizes the include and exclude expressions along with any test_module
        provided to collect the information about the test references and the associated test packages that will
        be involved in test run.
    """

    def __init__(self, root: str, excludes: Sequence[str], metafilters: Sequence[MetaFilter], 
                 method_prefix: str = "test", test_module: ModuleType = None):
        """
            Initializes a :class:`TestCollector` instance in order to process the test tree and
            collect references and test packages to be run.

            :param root: The root directory to scan for included tests
            :param excludes: A list or sequence of exclude expressions to apply during test collection operations.
            :param metafilters: A list of metadata filters to apply during test collection operations.
            :param module_prefix: The prefix or word that test methods will start with.  The default is 'test'.
            :param test_module: A test module which is passed for the debug workflow where a test module is run directly
                                as a script using the generic_test_entrypoint or in the debugger by Right-Click

        """
        self._root = root
        self._root_directory_listing = None
        self._excludes = excludes
        self._metafilters = metafilters
        self._method_prefix = method_prefix
        self._test_module = test_module
        self._test_references: Dict[str, TestRef] = {}
        self._import_errors = {}
        self._excluded_tests = []
        self._excluded_files = []
        self._excluded_path_prefixes, self._excluded_specifically = catagorize_exclusions(self._excludes)
        self._test_tree = None
        return

    @property
    def import_errors(self):
        """
            A list of import errors that were encountered while collecting test references.
        """
        return self._import_errors

    @property
    def references(self) -> Dict[str, TestRef]:
        """
            A list of :class:`TestReferences` that were collected.
        """
        return self._test_references

    @property
    def test_tree(self) -> TestGroup:
        return self._test_tree

    def collect_integrations_and_scopes(self) -> List[IntegrationCoupling]:
        """
            Iterates through all of the test references and collects the IntegrationCoupling(s) and
            ScopeCouplings that are found.  The collected IntegrationCoupling(s) and ScopeCoupling(s) can
            be used by integrations and scopes to participate in test framework startup.
        """

        referenced_integrations = {}
        referenced_scopes = {}

        # All the implicit registration of all the integration, scope and resource parameters sources
        # should have happend by now.  The ResourceRegistry should have a partially populated scope
        # tree.
        resource_registry.finalize_startup(self._test_references)

        unknown_parameter = resource_registry.unknown_parameters

        if len(unknown_parameter) > 0:
            err_msg_lines = [
                "TestCollector: Unable to resolve the following parameters.",
                "Missing Parameters:"
            ]
            subscriber_key_list = [upn for upn in unknown_parameter.keys()]
            subscriber_key_list.sort()
            for subscriber in subscriber_key_list:
                missing_params = unknown_parameter[subscriber]
                err_msg_lines.append("    {}".format(subscriber))
                for mparam in missing_params:
                    err_msg_lines.append("        {}".format(mparam))
            err_msg = os.linesep.join(err_msg_lines)
            raise AKitUnknownParameterError(err_msg) from None

        referenced_integrations = resource_registry.referenced_integrations
        referenced_scopes = resource_registry.referenced_scopes

        return referenced_integrations, referenced_scopes

    def collect_references(self, expression: str):
        """
            Collects and appends the test references based on the expression provided and the excludes
            for this class.  The `collect_references` method is intended to be called multiple times,
            once with each include expression provided by the users.  The :class:`TestCollector` will
            extend its collection of reference with each successive call.

            :param expression: An include expression to process and collect references for.
        """

        expr_package, expr_module, expr_testclass, expr_testname = parse_test_include_expression(expression, self._test_module, self._method_prefix)

        if expr_testclass != None:
            raise AKitSemanticError("TestPlus style tests do not support tests inside of classes.") from None

        # Find all the files that are included based on the expr_package, expr_module expressions
        included_files = []
        excluded_files = []

        if expr_package is not None or expr_module is not None:
            included_files, excluded_files = find_included_modules_under_root(self._root,
                expr_package, expr_module, excluded_path_prefixes=self._excluded_path_prefixes)

        self._excluded_files.extend(excluded_files)

        test_references, import_errors = collect_test_references(
            self._root, included_files, expr_package, expr_module, expr_testname, self._method_prefix)

        self._test_references.update(test_references)

        self._import_errors.update(import_errors)

        for modname, ifile, errmsg in import_errors.values():
            logger.error("TestCase: Import error filename=%r" % ifile)

        return

    def finalize_collection(self):
        """
            Applies the metadata filters and then creates the test tree.
        """
        
        testref_key_list = [k for k in self._test_references.keys()]

        temp_tree = TestGroup("<session>")
        for _, trval in self._test_references.items():
            temp_tree.add_descendent(trval)

        # Go through all of the test references and have all of the them
        # resolve their metadata reference chains
        temp_tree.resolve_metadata()

        # Apply metafilters to remove any tests that don't meet any specified
        # meta criteria
        if self._metafilters is not None and len(self._metafilters) > 0:
            filtered_references = {}

            for tref_key in testref_key_list:
                tref_obj = self._test_references[tref_key]
                if tref_obj.is_member_of_metaset(self._metafilters):
                    filtered_references[tref_key] = tref_obj

            del self._test_references
            self._test_references = filtered_references

        self._test_tree = TestGroup("<session>")
        for _, trval in self._test_references.items():
            self._test_tree.add_descendent(trval)

        return


"""
.. module:: console
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that is utilized by interactive consoles to activate the environment
               with logging to the console minimized in order to provide a good interactive
               console work experience.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os
import tempfile

from logging.handlers import RotatingFileHandler

from akit.exceptions import AKitSemanticError
from akit.environment.variables import ActivationProfile, AKIT_VARIABLES

__activation_profile__ = ActivationProfile.Console

# Guard against attemps to activate more than one, activation profile.
if AKIT_VARIABLES.AKIT_ACTIVATION_PROFILE is not None:
    errmsg = "An attempt was made to activate multiple environment activation profiles. profile={}".format(
        AKIT_VARIABLES.AKIT_ACTIVATION_PROFILE
    )
    raise AKitSemanticError(errmsg)

AKIT_VARIABLES.AKIT_ACTIVATION_PROFILE = ActivationProfile.Console

if AKIT_VARIABLES.AKIT_INTERACTIVE_CONSOLE:
    # If we are running in an interactive console, then we need to reduce the
    # console log level and we need to output log data to a console log file.

    temp_output_dir = tempfile.mkdtemp()

    # Only set the log levels if they were not previously set.  An option to a base
    # command may have set this in order to turn on a different level of verbosity
    if AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE is None:
        AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE = "QUIET"

    AKIT_VARIABLES.AKIT_JOBTYPE = "console"
    AKIT_VARIABLES.AKIT_OUTPUT_DIRECTORY = temp_output_dir

    # For console activation we don't want to log to the console and we want
    # to point the logs to a different output folder
    os.environ["AKIT_LOG_LEVEL_CONSOLE"] = AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE
    os.environ["AKIT_JOBTYPE"] = AKIT_VARIABLES.AKIT_JOBTYPE
    os.environ["AKIT_OUTPUT_DIRECTORY"] = AKIT_VARIABLES.AKIT_OUTPUT_DIRECTORY

    import akit.activation.base # pylint: disable=unused-import,wrong-import-position

    from akit.xlogging.foundations import logging_initialize, LoggingDefaults # pylint: disable=wrong-import-position

    LoggingDefaults.DefaultFileLoggingHandler = RotatingFileHandler
    logging_initialize()

    def showlog():

        print("OUTPUT FOLDER: {}".format(AKIT_VARIABLES.AKIT_OUTPUT_DIRECTORY))
        print("")

        return

else:

    # For console activation we don't want to log to the console and we want
    # to point the logs to a different output folder
    os.environ["AKIT_JOBTYPE"] = "console"

    import akit.activation.base # pylint: disable=unused-import,wrong-import-position

    from akit.xlogging.foundations import logging_initialize, LoggingDefaults # pylint: disable=wrong-import-position

    LoggingDefaults.DefaultFileLoggingHandler = RotatingFileHandler
    logging_initialize()



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import FrozenSet, List, Union, Optional

import bisect

from enum import Enum

from akit.exceptions import AKitSemanticError

class FeatureTagNodeMeta(type):

    def __new__(metacls, name, bases, namespace, **kwargs):
        cls = super().__new__(metacls, name, bases, namespace, **kwargs)
        cls.ID = cls.__qualname__.lower().replace(".", "/")
        return cls
    
    def __repr__(self):
        return "'{}'".format(self.ID)

class FeatureTag(metaclass=FeatureTagNodeMeta):
    ID = None

class FeatureAttachedObject:

    def __init__(self):

        self._feature_tags = []
        return

    @property
    def feature_tags(self) -> FrozenSet[str]:
        return frozenset(self._feature_tags)

    def extend_features(self, features_to_add: Union[List[FeatureTag], List[str]]):
        """
            Used by derived class and mixins to extend the feature tags associated with
            a feature attached object.
        """

        if len(features_to_add) > 0:
            first_item = features_to_add[0]
            if isinstance(first_item, str):
                # We insert the features into the list sorted so we can make finding
                # features faster.
                for ft in features_to_add:
                    bisect.insort(self._feature_tags, ft)
            elif issubclass(first_item, FeatureTag):
                # We insert the features into the list sorted so we can make finding
                # features faster.
                for ft in features_to_add:
                    bisect.insort(self._feature_tags, ft.ID)
            else:
                errmsg = "The 'features_to_add' parameter must contain items of type 'FeatureTag' or 'str'. item={}".format(
                    repr(first_item)
                )
                raise AKitSemanticError(errmsg)

        return
    
    def has_all_features(self, feature_list: Union[List[FeatureTag], List[str]]):

        has_all = True

        if len(feature_list) == 0:
            errmsg = "has_all_features: 'feature_list' cannot be empty."
            raise AKitSemanticError(errmsg)

        first_item = feature_list[0]
        if isinstance(first_item, str):
            for feature in feature_list:
                fid = feature
                hasfeature = fid in self._feature_tags
                if not hasfeature:
                    has_all = False
                    break
        elif issubclass(first_item, FeatureTag):
            for feature in feature_list:
                fid = feature.ID
                hasfeature = fid in self._feature_tags
                if not hasfeature:
                    has_all = False
                    break
        else:
            errmsg = "The 'feature_list' parameter must contain items of type 'FeatureTag' or 'str'. item={}".format(
                repr(first_item)
            )
            raise AKitSemanticError(errmsg)

        return has_all

    def has_any_feature(self, feature_list: List[FeatureTag]):

        has_any = False

        if len(feature_list) == 0:
            errmsg = "has_all_features: 'feature_list' cannot be empty."
            raise AKitSemanticError(errmsg)

        first_item = feature_list[0]
        if isinstance(first_item, str):
            for feature in feature_list:
                fid = feature

                hasfeature = fid in self._feature_tags
                if hasfeature:
                    has_any = True
                    break
        elif issubclass(first_item, FeatureTag):
            for feature in feature_list:
                fid = feature.ID

                hasfeature = fid in self._feature_tags
                if hasfeature:
                    has_any = True
                    break
        else:
            errmsg = "The 'feature_list' parameter must contain items of type 'FeatureTag' or 'str'. item={}".format(
                repr(first_item)
            )
            raise AKitSemanticError(errmsg)

        return has_any

    def has_feature(self, feature: Union[FeatureTag, str]):
        fid = None

        if isinstance(feature, str):
            fid = feature
        elif issubclass(feature, FeatureTag):
            fid = feature.ID
        else:
            errmsg = "The 'feature' parameter must be of type 'FeatureTag' or 'str'. item={}".format(
                repr(feature)
            )
            raise AKitSemanticError(errmsg)

        hasfeature = fid in self._feature_tags

        return hasfeature


class FeatureMask(dict):

    def __init__(self, *, required_features: Optional[List[FeatureTag]]=None, excluded_features: Optional[List[FeatureTag]]=None, **kwargs):
        super().__init__(required_features=required_features, excluded_features=excluded_features, **kwargs)
        return


class FeatureFilter:

    def __init__(self, *, required_features: Optional[List[FeatureTag]]=None, excluded_features: Optional[List[FeatureTag]]=None, **kwargs):
        self._required_features = required_features
        self._excluded_features = excluded_features
        return

    def filter(self, device_list: List[FeatureAttachedObject]) -> List[FeatureAttachedObject]:

        matching_devices = []

        for fd in device_list:
            if self._required_features is not None:
                has_req_features = fd.has_all_features(self._required_features)
            else:
                has_req_features = True

            if self._excluded_features is not None:
                has_excl_features = fd.has_any_feature(self._excluded_features)
            else:
                has_excl_features = False

            if has_req_features and not has_excl_features:
                matching_devices.append(fd)

        return matching_devices




"""
.. module:: upnpgenerator
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that processes service descriptions and generates proxy service wrappers used
               to communicated with UPNP services.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional, Tuple, Union

import io
import os
import sys

from argparse import ArgumentParser, ArgumentError

from xml.etree.ElementTree import Element
from xml.etree.ElementTree import fromstring as xml_fromstring

from akit.paths import ensure_directory_is_package

from akit.interop.coordinators.upnpcoordinator import UpnpCoordinator

# pylint: disable=unused-import

UPNP_SERVICE_NAMESPACE = "urn:schemas-upnp-org:service-1-0"

MANUFACTURER_TO_BASE_CLASSES = {
    'SonosInc': ("akit.interop.upnp.devices.embedded.upnpdevice", "SonosDevice")
}

CONTENT_PROXY_FILE_HEADER = """
    NOTE: This is a code generated file.  This file should not be edited directly.
"""

PROXY_BASE_CLASS_IMPORT = "from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy"
PROXY_BASE_CLASS_NAME = "UpnpServiceProxy"

TEMPLATE_CLASS_PREFIX = """

from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
%(base_class_import)s

class %(class_name)s(%(base_class_name)s, LoadableExtension):
    \"""
        This is a code generated proxy class to the '%(service_name)s' service.
    \"""

    SERVICE_MANUFACTURER = '%(service_manufacturer)s'
    SERVICE_TYPE = '%(service_type)s'

    SERVICE_DEFAULT_VARIABLES = {%(svc_default_vars)s}

    SERVICE_EVENT_VARIABLES = {%(svc_event_vars)s}
"""

TEMPLATE_ACTION_NO_RETURN = """
    def action_%(action_name)s(self%(in_params_comma)s%(in_params_list)s, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        \"""
            Calls the %(action_name)s action.
        \"""
        arguments = %(args_dict)s

        self.call_action("%(action_name)s", arguments=arguments, aspects=aspects)

        return
"""

TEMPLATE_ACTION_WITH_RETURN = """
    def action_%(action_name)s(self%(in_params_comma)s%(in_params_list)s, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        \"""
            Calls the %(action_name)s action.

            :returns: %(out_params_list)s
        \"""
        arguments = %(args_dict)s

        out_params = self.call_action("%(action_name)s", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in (%(out_params_list)s,)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args
"""

def node_lower_strip_text(txt: Union[str, None]) -> Union[str, None]:
    """
        Converts text to lower case and strips leading and trailing whitespace if not None.

        :param txt: The text to convert to lower case and to strip.
    """
    if txt is not None:
        txt = txt.lower().strip()
    return txt

def node_strip_text(txt: Union[str, None]) -> Union[str, None]:
    """
        Strips leading and trailing whitespace from the given text if not None.

        :param txt: The text to strip.
    """
    if txt is not None:
        txt = txt.strip()
    return txt

def generate_upnp_service_proxy(serviceManufacturer: str, serviceType: str, serviceName: str, className: str,
                                variablesTable: dict,typesTable: dict, eventsTable: dict, actionsTable: dict,
                                base_class_name: str=PROXY_BASE_CLASS_NAME,
                                base_class_import: str=PROXY_BASE_CLASS_IMPORT) -> str:
    """
        Generates a service proxy using the parameters provided.

        :param servicesDir: The directory to output the service proxy to.
        :param serviceManufacturer: The name of the manufacturer for the device that a service proxy is being generated for.
        :param serviceName: The name of the service the service proxy is being generated for.
        :param variablesTable: A table containing the information about the variables associated with a service.
        :param typesTable: A table containing the types associated with the variables used by the service.
        :param eventsTable: A table containing information about the events published by the services.
        :param actionsTable: A table containing information about the actions that can be called on the service.

    """
    # pylint: disable=unused-argument

    service_variables_content = ""

    class_fill_dict = {
        "base_class_name": base_class_name,
        "base_class_import": base_class_import,
        "class_name": className,
        "service_name": serviceName,
        "service_manufacturer": serviceManufacturer,
        "service_type": serviceType,
        "service_variables": service_variables_content
    }

    variable_names_sorted = [ k for k in variablesTable.keys() ]
    variable_names_sorted.sort()

    event_variable_lines = []
    default_variable_lines = []

    for var_name in variable_names_sorted:
        variable_info = variablesTable[var_name]

        var_name = variable_info["name"]
        var_type = variable_info["dataType"]
        var_send_events = variable_info["sendEvents"]

        var_allowed_list = None
        if "allowedValueList" in variable_info:
            var_allowed_list = variable_info["allowedValueList"]

        var_default_value = None
        if "defaultValue" in variable_info:
            var_default_value = variable_info["defaultValue"]

        default_entry = ", \"default\": None"
        if var_default_value is not None:
            default_entry = ", \"default\": \"%s\"" % var_default_value

        allowed_list_entry = ", \"allowed_list\": None"
        if var_allowed_list is not None:
            allowed_list_entry = ", \"allowed_list\": \"%s\"" % var_allowed_list

        if var_send_events == 'yes':
            table_var_line = "\"%s\": { \"data_type\": \"%s\"%s%s},\n" % (var_name, var_type, default_entry, allowed_list_entry)
            event_variable_lines.append(table_var_line)
        else:
            table_var_line = "\"%s\": { \"data_type\": \"%s\"%s%s},\n" % (var_name, var_type, default_entry, allowed_list_entry)
            default_variable_lines.append(table_var_line)

    if len(event_variable_lines) > 0:
        service_variables_content = "\n        " + "        ".join(event_variable_lines) + "    "
        class_fill_dict["svc_event_vars"] = service_variables_content
    else:
        class_fill_dict["svc_event_vars"] = ""

    if len(default_variable_lines) > 0:
        service_variables_content = "\n        " + "        ".join(default_variable_lines) + "    "
        class_fill_dict["svc_default_vars"] = service_variables_content
    else:
        class_fill_dict["svc_default_vars"] = ""

    spf = io.StringIO()
    spf.write('"""\n')
    spf.write(CONTENT_PROXY_FILE_HEADER)
    spf.write('"""\n')
    spf.write('\n')
    spf.write(TEMPLATE_CLASS_PREFIX % class_fill_dict)

    action_names_sorted = [ k for k in actionsTable.keys() ]
    action_names_sorted.sort()

    for action_name in action_names_sorted:

        action_info = actionsTable[action_name]

        in_params_list = ""
        out_params_list = ""

        args_dict = "{ }"
        args_in_keys = action_info["args_in_keys"]
        if len(args_in_keys) > 0:
            in_params_list = ", ".join(args_in_keys)
            args_dict = "{\n"
            for arg_key in args_in_keys:
                args_dict += '            "%s": %s,\n' % (arg_key, arg_key)
            args_dict += "        }"

        in_params_comma = ""
        if len(in_params_list) > 0:
            in_params_comma = ", "

        args_out_keys = [ '"%s"' % ok for ok in action_info["args_out_keys"] ]
        if len(args_out_keys) > 0:
            out_params_list = ", ".join(args_out_keys)

        action_fill = {
            "action_name": action_name,
            "in_params_list": in_params_list,
            "in_params_comma": in_params_comma,
            "out_params_list": out_params_list,
            "args_dict": args_dict
        }
        if len(out_params_list) > 0:
            spf.write(TEMPLATE_ACTION_WITH_RETURN % action_fill)
        else:
            spf.write(TEMPLATE_ACTION_NO_RETURN % action_fill)

    return spf.getvalue()

def process_action_list(svcActionListNode: Element, namespaces: Optional[dict] = None) -> dict:
    """
        Processes the action list node for a service description and creates a table with all the actions available
        on the service.

        :param svcActionListNode: The action list xml node from the serivce description document.
        :param namespace: The namespace to use when processing the XML dom nodes.

        :returns: Returns a table of actions supported by the service.
    """
    actionsTable = {}

    actionNodeList = svcActionListNode.findall("action", namespaces=namespaces)
    for actionNode in actionNodeList:
        action_name = node_strip_text(actionNode.find("name", namespaces=namespaces).text)
        args_in_keys = []
        args_in_table = {}
        args_out_keys = []
        args_out_table = {}

        argumentListNode = actionNode.find("argumentList", namespaces=namespaces)
        if argumentListNode is not None:
            argumentNodeList = argumentListNode.findall("argument", namespaces=namespaces)
            for argumentNode in argumentNodeList:
                arg_info = {}
                arg_name = node_strip_text(argumentNode.find("name", namespaces=namespaces).text)
                arg_direction = node_lower_strip_text(argumentNode.find("direction", namespaces=namespaces).text)

                arg_related = None
                argRelatedNode = argumentNode.find("relatedStateVariable", namespaces=namespaces)
                if argRelatedNode is not None:
                    arg_related = node_strip_text(argRelatedNode.text)

                arg_info["name"] = arg_name
                arg_info["direction"] = arg_direction
                arg_info["relatedStateVariable"] = arg_related

                if arg_direction == "in":
                    args_in_keys.append(arg_name)
                    args_in_table[arg_name] = arg_info
                elif arg_direction == "out":
                    args_out_keys.append(arg_name)
                    args_out_table[arg_name] = arg_info
                else:
                    raise ValueError("Invalid argument direction %s" % arg_direction)

        action_info = {
            "name": action_name,
            "args_in": args_in_table,
            "args_in_keys": args_in_keys,
            "args_out": args_out_table,
            "args_out_keys": args_out_keys
        }

        actionsTable[action_name] = action_info

    return actionsTable

def process_service_state_table(svcStateTableNode: Element, namespaces: Optional[dict] = None) -> Tuple[dict, dict, dict]:
    """
        Processes the action list node for a service description and creates a table with all the actions available
        on the service.

        :param svcStateTableNode: The state table XML node from the serivce description document.
        :param namespace: The namespace to use when processing the XML dom nodes.

        :returns: Returns tables for the variables, types and events supported by the service.
    """
    variablesTable = {}
    typesTable = {}
    eventsTable = {}

    stateVariableNodeList = svcStateTableNode.findall("stateVariable", namespaces=namespaces)

    for stateVariableNode in stateVariableNodeList:

        variable_info = {}

        var_name = node_strip_text(stateVariableNode.find("name", namespaces=namespaces).text)
        variable_info["name"] = var_name

        send_events = "no"
        if "sendEvents" in stateVariableNode.attrib:
            send_events = node_lower_strip_text(stateVariableNode.attrib["sendEvents"])
        else:
            sendEventsNode = stateVariableNode.find("sendEventsAttribute")
            if sendEventsNode is not None:
                send_events = node_lower_strip_text(sendEventsNode.text)

        variable_info["sendEvents"] = send_events


        var_type = stateVariableNode.find("dataType", namespaces=namespaces).text.strip()
        variable_info["dataType"] = var_type

        allowedValueListNode = stateVariableNode.find("allowedValueList", namespaces=namespaces)
        if allowedValueListNode is not None:
            allowed_value_list = []
            for allowedValueNode in allowedValueListNode.findall("allowedValue", namespaces=namespaces):
                allowed_value = node_strip_text(allowedValueNode.text)
                allowed_value_list.append(allowed_value)
            variable_info["allowedValueList"] = allowed_value_list

        defaultValueNode = stateVariableNode.find("defaultValue", namespaces=namespaces)
        if defaultValueNode is not None:
            default_value = node_strip_text(defaultValueNode.text)
            variable_info["defaultValue"] = default_value

        if var_name.startswith("A_ARG_TYPE_"):
            typesTable[var_name] = variable_info
        else:
            variablesTable[var_name] = variable_info
            if send_events == "yes":
                eventsTable[var_name] = variable_info

    return variablesTable, typesTable, eventsTable


def generate_service_proxies(svc_desc_directory: str, svc_proxy_directory: str):
    """
        Processes the XML service description documents in the description documents folder and generates the
        service proxy modules.  Then outputs the generated proxy modules to the service proxy foloder specified.

        :param svc_desc_directory: The directory that contains the service description documents to process.
        :param svc_proxy_directory: The directory that is the output directory for the service proxy modules.
    """
    for dirpath, _, filenames in os.walk(svc_desc_directory, topdown=True):
        for nxtfile in filenames:

            serviceType, nxtfile_ext = os.path.splitext(nxtfile)
            if nxtfile_ext != ".xml":
                continue

            serviceManufacturer = os.path.basename(dirpath)

            svc_content = None

            fullpath = os.path.join(dirpath, nxtfile)
            with open(fullpath, 'r') as xf:
                svc_content = xf.read()

            docNode = xml_fromstring(svc_content)
            if docNode is not None:

                namespaces = None
                doc_node_tag = docNode.tag
                if doc_node_tag.find("}") > 0:
                    default_ns = doc_node_tag[doc_node_tag.find("{") + 1:doc_node_tag.find("}")]
                    namespaces = {"": default_ns}

                variablesTable = {}
                typesTable = {}
                eventsTable = {}

                svcStateTableNode = docNode.find("serviceStateTable", namespaces=namespaces)
                if svcStateTableNode is not None:
                    variablesTable, typesTable, eventsTable = process_service_state_table(svcStateTableNode, namespaces=namespaces)

                if serviceType.find("DeviceProperties") > 0:
                    print("found")

                actionsTable = {}
                actionListNode = docNode.find("actionList", namespaces=namespaces)
                if actionListNode is not None:
                    actionsTable = process_action_list(actionListNode, namespaces=namespaces)

                ensure_directory_is_package(svc_proxy_directory, package_title="Services directory module")

                manufacturerDir = os.path.join(svc_proxy_directory, serviceManufacturer)
                if not os.path.exists(manufacturerDir):
                    os.makedirs(manufacturerDir)

                service_type_parts = serviceType.split(":")
                serviceName = service_type_parts[3] + service_type_parts[-1]

                className = serviceName + "ServiceProxy"
                file_base = className.lower() + ".py"

                dest_file_full = os.path.join(manufacturerDir, file_base)
                content = generate_upnp_service_proxy(serviceManufacturer, serviceName, serviceType,
                        className, variablesTable, typesTable, eventsTable, actionsTable)

                with open(dest_file_full, 'w') as df:
                    df.write(content)

            else:
                errmsg = "WARNING: No serice node found in file:\n    %s\n" % fullpath
                print(errmsg, file=sys.stderr)

    return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

HELP_ROOT = "The root directory to use when scanning for tests."
HELP_EXCLUDES = "Add a test inclusion expression."
HELP_INCLUDES = "Add a test exclusion expression."
HELP_OUTPUT = "The output directory where results and artifacts are collected."
HELP_START = r"A time stamp to associate with the start of the run. Example: 2020-10-17T15:30:11.989120  Bash: date +%Y-%m-%dT%H:%M:%S.%N"
HELP_BRANCH = "The name of the branch to associate with the test run results."
HELP_BUILD = "The name of the build to associate with the test run results."
HELP_FLAVOR = "The name of the flavor to associate with the test run results."
HELP_JOB_INITIATOR = "The name of the initiator of the job."
HELP_JOB_LABEL = "The label associated with the job."
HELP_JOB_NAME = "The name of the job."
HELP_JOB_OWNER = "The ownerid and possibly appended ';(display name)' of the test run result owner."
HELP_CREDENTIALS = "The full path of the credentials file to use for the testrun."
HELP_LANDSCAPE = "The full path of the landscape file to use for the testrun."
HELP_LANDSCAPE_NAME = "The name of the landscape to use when looking for the landscape file."
HELP_RUNTIME = "The full path of the runtime file to use for the testrun."
HELP_RUNTIME_NAME = "The name of the runtime to use when looking for the runtime file."
HELP_TOPOLOGY = "The full path of the topology file to use for the testrun."
HELP_TOPOLOGY_NAME = "The name of the topology to use when looking for the topology file."
HELP_RUNID = "A uuid to use for the run id for the testrun."
HELP_CONSOLE_LOG_LEVEL = "The logging level for console output."
HELP_FILE_LOG_LEVEL = "The logging level for logfile output."
HELP_DEBUG = "Output debug information to the console."
HELP_DEBUGGER = "Debugger to active during the test run."
HELP_BREAKPOINT = "The breakpoint to activate for the test run."
HELP_TIMETRAVEL = "Enables tracing for Time-Travel-Debugging."
HELP_TIMEPORTAL = "The name of a time portal to open for Time-Travel-Debugging."
HELP_PRERUN_DIAGNOSTIC = "Flag indicating that a pre run diagnostic should be performed."
HELP_POSTRUN_DIAGNOSTIC = "Flag indicating that a post run diagnostic should be performed."
HELP_INCLUDE_MARKER_EXP = "Allows the specification of a marker expression that if matched will include a test."
HELP_EXCLUDE_MARKER_EXP = "Allows the specification of a marker expression that if matched will exclude a test."



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Union

from collections import defaultdict

import yaml

from xml.etree.ElementTree import fromstring as xml_from_string

def etree_to_dict(cursor) -> dict:
    """
       Converts an ElementTree node and its descendant children to a dictionary
       object.

       :param cursor: The root element tree node to convert to a dictionary 
    """

    info = {cursor.tag: {} if cursor.attrib else None}

    children = list(cursor)
    if children:
        dd = defaultdict(list)
        for dc in map(etree_to_dict, children):
            for k, v in dc.items():
                dd[k].append(v)
        info = {cursor.tag: {k: v[0] if len(v) == 1 else v
                     for k, v in dd.items()}}

    if cursor.attrib:
        info[cursor.tag].update(('@' + k, v)
                        for k, v in cursor.attrib.items())

    if cursor.text:
        text = cursor.text.strip()
        if children or cursor.attrib:
            if text:
                info[cursor.tag]['#text'] = text
        else:
            info[cursor.tag] = text

    return info

def parse_xml_to_dict(xml_content) -> dict:
    """
        Processes xml content and returns a dictionary based on the xml content.

        :param xml_content: The xml document content to process and convert.
    """
    rtn_info = None
    troot = xml_from_string(xml_content)
    rtn_info = etree_to_dict(troot)

    return rtn_info

def format_xml_as_yaml(xml_content: str, split_lines: bool = False) -> Union[str, List[str]]:
    """
        Processes xml content and converts it to YAML

        :param xml_content: The xml document content to process and convert.
        :param split_lines: A boolean indicating that the yaml content should be
                            returned as a multi-line list of strings.
    """

    dict_info = parse_xml_to_dict(xml_content)

    content = yaml.dump(dict_info)

    if split_lines:
        content = content.splitlines(False)

    return content



"""
.. module:: pkey
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module that contains methods for detecting the type of a private key file and
               for working with private key files.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

PUB_KEY_BEGIN_RSA_PKCS1 = "-----BEGIN RSA PUBLIC KEY-----"
PUB_KEY_END_RSA_PKCS1 = "-----END RSA PUBLIC KEY-----"

PRV_KEY_BEGIN_RSA_PKCS1 = "-----BEGIN RSA PRIVATE KEY-----"
PRV_KEY_END_RSA_PKCS1 = "-----END RSA PRIVATE KEY-----"

PUB_KEY_BEGIN_RSA_PKCS8 = "-----BEGIN PUBLIC KEY-----"
PUB_KEY_END_RSA_PKCS8 = "-----END PUBLIC KEY-----"

PRV_KEY_BEGIN_RSA_PKCS8 = "-----BEGIN PRIVATE KEY-----"
PRV_KEY_END_RSA_PKCS8 = "-----END PRIVATE KEY-----"

PRV_KEYENC_BEGIN_RSA_PKCS8 = "-----BEGIN ENCRYPTED PRIVATE KEY-----"
PRV_KEYENC_END_RSA_PKCS8 = "-----END ENCRYPTED PRIVATE KEY-----"

# ASCII magic "openssh-key-v1" plus null byte
"6f70656e7373682d6b65792d7631 00"

def detect_private_key_type(filename: str) -> str:
    """
        Inspects the private key file specified and determines the keytype found in the file.

        :param filename: The filename of a private key file.

        :returns: A string for the keytype
    """
    # pylint: disable=invalid-name,no-else-break

    with open(filename, 'r') as kf:
        lines = kf.readlines()

    keytype = None
    for ln in lines:
        if ln.find(PRV_KEY_BEGIN_RSA_PKCS1) > -1:
            keytype = "rsa-pkcs1"
            break
        elif ln.find(PRV_KEY_BEGIN_RSA_PKCS8) > -1:
            keytype = "rsa-pkcs8"
            break
        elif ln.find(PRV_KEYENC_BEGIN_RSA_PKCS8) > -1:
            keytype = "rsa-pkcs8-enc"
            break

    return keytype



"""
.. module:: upnpdevice1
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the XML classes used to process the XML content in a UpnpDevice1
               description document.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional, Union

import json

from xml.etree.ElementTree import tostring as xml_tostring
from xml.etree.ElementTree import register_namespace
from xml.etree.ElementTree import Element

from akit.exceptions import AKitNotOverloadedError


UPNP_DEVICE1_NAMESPACE = "urn:schemas-upnp-org:device-1-0"
UPNP_DEVICE1_TAG_PREFIX = "{urn:schemas-upnp-org:device-1-0}"


class UpnpDevice1ElementTags:
    """
        Contains constanst for the UpnpDevice1 XML element tag names.
    """
    deviceType = UPNP_DEVICE1_TAG_PREFIX + "deviceType"
    friendlyName = UPNP_DEVICE1_TAG_PREFIX + "friendlyName"
    manufacturer = UPNP_DEVICE1_TAG_PREFIX + "manufacturer"
    manufacturerURL = UPNP_DEVICE1_TAG_PREFIX + "manufacturerURL"
    modelDescription = UPNP_DEVICE1_TAG_PREFIX + "modelDescription"
    modelName = UPNP_DEVICE1_TAG_PREFIX + "modelName"
    modelNumber = UPNP_DEVICE1_TAG_PREFIX + "modelNumber"
    modelURL = UPNP_DEVICE1_TAG_PREFIX + "modelURL"
    serialNumber = UPNP_DEVICE1_TAG_PREFIX + "serialNumber"
    UDN = UPNP_DEVICE1_TAG_PREFIX + "UDN"
    UPC = UPNP_DEVICE1_TAG_PREFIX + "UPC"
    presentationURL = UPNP_DEVICE1_TAG_PREFIX + "presentationURL"
    deviceList = UPNP_DEVICE1_TAG_PREFIX + "deviceList"
    iconList = UPNP_DEVICE1_TAG_PREFIX + "iconList"
    serviceList = UPNP_DEVICE1_TAG_PREFIX + "serviceList"


class UpnpDevice1ElementBase:
    """
        The :class:`UpnpDevice1ElementBase` object provides a place to hang common data and
        methods for elements from the UpnpDevice1 document elements.
    """

    def __init__(self, ref_node: Element, namespaces: Optional[dict] = None):
        """
            Constructs a :class:`UpnpDevice1ElementBase` derived object initializing the
            reference node and namespace dictionary used for XML document processing.

            :param ref_node: The XML :class:`Element` node associated from the device description.
            :param namespaces: A dictionary of namespaces to use when processing the XML elements.
        """
        self._ref_node = ref_node
        self._namespaces = namespaces
        return

    @property
    def xmlNode(self) -> Element:
        """
            The XML node associated with the icon information.
        """
        return self._ref_node

    def to_dict(self, brief=False) -> dict:
        """
            Converts the icon data to a python dictionary.
        """
        # pylint: disable=no-self-use, unused-argument
        raise AKitNotOverloadedError("UpnpDevice1Base.to_dict: Must be overridden by derived classes.") from None

    def to_json(self) -> str:
        """
            Converts the icon data to a dictionary and then converts it to a JSON string.
        """
        dval = self.to_dict()
        json_str = json.dumps(dval, indent=4)
        return json_str

    def _find_value(self, path, namespaces) -> Union[str, None]:
        """
            Lookup a node using the path specified and then get its text value.

            :param path: The path to the node that is being found.
            :param namespaces: A dictionary of namespaces to use when processing the XML elements.

            :returns: The text value of the node associated with the specified path or None.
        """
        # pylint: disable=broad-except
        rtnval = None

        try:
            valNode = self._ref_node.find(path, namespaces=namespaces)
            if valNode is not None:
                rtnval = valNode.text
        except Exception as err:
            print(str(err))
            register_namespace('', None)
            xmlcontent = xml_tostring(self._ref_node)
            print(xmlcontent)
            print("")

        return rtnval



class UpnpDevice1Icon(UpnpDevice1ElementBase):
    """
        The :class:`UpnpDevice1Icon` object is used to process the icon description portion of a device
        description document.

        .. code-block:: xml

            <icon>
                <id>0</id>
                <mimetype>image/png</mimetype>
                <width>48</width>
                <height>48</height>
                <depth>24</depth>
                <url>/img/icon-S18.png</url>
            </icon>

    """

    def __init__(self, iconNode: Element, namespaces: Optional[dict] = None):
        """
            Creates a :class:`UpnpDevice1Icon` object to provide a shell for processing
            `icon` element information from a device description document.

            :param iconNode: The XML :class:`Element` node associated from the icon description.
            :param namespaces: A dictionary of namespaces to use when processing the XML elements.
        """
        super(UpnpDevice1Icon, self).__init__(iconNode, namespaces=namespaces)
        return

    @property
    def depth(self) -> str:
        """
            The color depth in bits of the the icon.
        """
        rtnval = self._find_value("depth", namespaces=self._namespaces)
        return rtnval

    @property
    def height(self) -> str:
        """
            The height of the icon in pixels.
        """
        rtnval = self._find_value("height", namespaces=self._namespaces)
        return rtnval

    @property
    def id(self) -> str:
        """
            The id that is assciated with the icon in the icon list.
        """
        rtnval = self._find_value("id", namespaces=self._namespaces)
        return rtnval

    @property
    def url(self) -> str:
        """
            The full or relative url of the icon.  If the URL does not have a scheme and host then
            it is likely a relative URL.
        """
        rtnval = self._find_value("url", namespaces=self._namespaces)
        return rtnval

    @property
    def mimetype(self) -> str:
        """
            The mime-type of the icon.
        """
        rtnval = self._find_value("mimetype", namespaces=self._namespaces)
        return rtnval

    @property
    def width(self) -> str:
        """
            The width of the icon in pixels.
        """
        rtnval = self._find_value("width", namespaces=self._namespaces)
        return rtnval

    def to_dict(self, brief=False) -> dict:
        """
            Converts the icon data to a python dictionary.
        """
        # pylint: disable=unused-argument
        dval = {
            "depth": self.depth,
            "height": self.height,
            "id": self.id,
            "url": self.url,
            "mimetype": self.mimetype,
            "width": self.width
        }
        return dval


class UpnpDevice1Service(UpnpDevice1ElementBase):
    """
        The :class:`UpnpDevice1Service` object is used to process the service description items in the
        service description list from a device description document.

        .. code-block:: xml

            <service>
                <serviceType>urn:schemas-upnp-org:service:AlarmClock:1</serviceType>
                <serviceId>urn:upnp-org:serviceId:AlarmClock</serviceId>
                <controlURL>/AlarmClock/Control</controlURL>
                <eventSubURL>/AlarmClock/Event</eventSubURL>
                <SCPDURL>/xml/AlarmClock1.xml</SCPDURL>
            </service>

    """
    def __init__(self, svcManufacturer: str, svcNode: Element, namespaces: Optional[dict] = None):
        """
            Creates a :class:`UpnpDevice1Service` object to provide a shell for processing
            `service` element information from a device description document.

            :param svcManufacturer: The manufacturer associated with the device and service descriptions.
            :param svcNode: The XML :class:`Element` node associated from the service description.
            :param namespaces: A dictionary of namespaces to use when processing the XML elements.
        """
        super(UpnpDevice1Service, self).__init__(svcNode, namespaces=namespaces)

        self._svcManufacturer = svcManufacturer
        return

    @property
    def controlURL(self) -> str:
        """
            Returns the control URL which is the URL for action calls.
        """
        rtnval = self._find_value("controlURL", namespaces=self._namespaces)
        return rtnval

    @property
    def eventSubURL(self) -> str:
        """
            Returns the event sub URL which is utilized to subscribe to events.
        """
        rtnval = self._find_value("eventSubURL", namespaces=self._namespaces)
        return rtnval

    @property
    def SCPDURL(self) -> str:
        """
            Returns the service description URL.
        """
        rtnval = self._find_value("SCPDURL", namespaces=self._namespaces)
        return rtnval

    @property
    def serviceId(self) -> str:
        """
            Returns the service ID of the service.
        """
        rtnval = self._find_value("serviceId", namespaces=self._namespaces)
        return rtnval

    @property
    def serviceManufacturer(self) -> str:
        """
            Returns the manufacturer name of the service.
        """
        return self._svcManufacturer

    @property
    def serviceType(self) -> str:
        """
            Returns the service type of the service.
        """
        rtnval = self._find_value("serviceType", namespaces=self._namespaces)
        return rtnval

    def to_dict(self, brief=False) -> dict:
        """
            Converts the icon data to a python dictionary.
        """
        # pylint: disable=unused-argument
        dval = {
            "controlURL": self.controlURL,
            "eventSubURL": self.eventSubURL,
            "SCPDURL": self.SCPDURL,
            "serviceId": self.serviceId,
            "serviceManufacturer": self.serviceManufacturer,
            "serviceType": self.serviceType
        }

        return dval

class UpnpDevice1SpecVersion(UpnpDevice1ElementBase):
    """
        The :class:`UpnpDevice1SpecVersion` object is used to process the service description items in the
        service description list from a device description document.

        .. code-block:: xml

            <specVersion>
                <major>1</major>
                <minor>0</minor>
            </specVersion>

    """
    def __init__(self, verNode: Element, namespaces: Optional[dict] = None):
        """
            Creates a :class:`UpnpDevice1SpecVersion` object to provide a shell for processing
            `specVersion` element information from a device description document.

            :param svcManufacturer: The manufacturer associated with the device and service descriptions.
            :param svcNode: The XML :class:`Element` node associated from the service description.
            :param namespaces: A dictionary of namespaces to use when processing the XML elements.
        """
        super(UpnpDevice1SpecVersion, self).__init__(verNode, namespaces=namespaces)
        return

    @property
    def major(self) -> str:
        """
            Returns the major version of the device description.
        """
        rtnval = self._find_value("major", namespaces=self._namespaces)
        return rtnval

    @property
    def minor(self) -> str:
        """
            Returns the minor version of the device description.
        """
        rtnval = self._find_value("minor", namespaces=self._namespaces)
        return rtnval

    def to_dict(self, brief=False) -> dict:
        """
            Converts the icon data to a python dictionary.
        """
        # pylint: disable=unused-argument
        dval = {
            "major": self.major,
            "minor": self.minor
        }

        return dval


class UpnpDevice1Device(UpnpDevice1ElementBase):
    """
        The UPNP Root device is the base device for the hierarchy that is
        associated with a unique network devices location.  The :class:`RootDevice`
        and its subdevices are linked by thier location url.

        http://www.upnp.org/specs/arch/UPnP-arch-DeviceArchitecture-v1.0.pdf

        .. code-block:: xml

            <device>
                <deviceType>urn:schemas-upnp-org:device:deviceType:v</deviceType>
                <friendlyName>short user-friendly title</friendlyName>
                <manufacturer>manufacturer name</manufacturer>
                <manufacturerURL>URL to manufacturer site</manufacturerURL>
                <modelDescription>long user-friendly title</modelDescription>
                <modelName>model name</modelName>
                <modelNumber>model number</modelNumber>
                <modelURL>URL to model site</modelURL>
                <serialNumber>manufacturer's serial number</serialNumber>
                <UDN>uuid:UUID</UDN>
                <UPC>Universal Product Code</UPC>
                <iconList>
                    <icon>
                    <mimetype>image/format</mimetype>
                    <width>horizontal pixels</width>
                    <height>vertical pixels</height>
                    <depth>color depth</depth>
                    <url>URL to icon</url>
                    </icon>
                    XML to declare other icons, if any, go here
                </iconList>
                <serviceList>
                    <service>
                        <serviceType>urn:schemas-upnp-org:service:serviceType:v</serviceType>
                        <serviceId>urn:upnp-org:serviceId:serviceID</serviceId>
                        <SCPDURL>URL to service description</SCPDURL>
                        <controlURL>URL for control</controlURL>
                        <eventSubURL>URL for eventing</eventSubURL>
                    </service>
                    Declarations for other services defined by a UPnP Forum working committee (if any)
                    go here Declarations for other services added by UPnP vendor (if any) go here
                </serviceList>
                <presentationURL>URL for presentation</presentationURL>
            </device>

    """

    def __init__(self, devNode: Element, namespaces: Optional[dict] = None):
        """
            Creates a :class:`UpnpDevice1Device` object to provide a shell for processing
            `device` element information from a device description document.

            :param devNode: The XML :class:`Element` node associated from the service description.
            :param namespaces: A dictionary of namespaces to use when processing the XML elements.
        """
        super(UpnpDevice1Device, self).__init__(devNode, namespaces=namespaces)
        return

    @property
    def deviceList(self) -> List['akit.interop.upnp.xml.upnpdevice1.UpnpDevice1Device']:
        """
            Returns a listing of any embedded device descriptions from the device description.
        """
        resultList = []
        listNode = self._ref_node.find("deviceList", namespaces=self._namespaces)
        if listNode is not None:
            resultList = [ UpnpDevice1Device(child, namespaces=self._namespaces) for child in listNode ]
        return resultList

    @property
    def iconList(self) -> List[UpnpDevice1Icon]:
        """
            Returns a listing of any icon descriptions from the device description.
        """
        resultList = []
        listNode = self._ref_node.find("iconList", namespaces=self._namespaces)
        if listNode is not None:
            resultList = [ UpnpDevice1Icon(child, namespaces=self._namespaces) for child in listNode ]
        return resultList

    @property
    def serviceList(self) -> List[UpnpDevice1Service]:
        """
            Returns a listing of any service descriptions from the device description.
        """
        manufacturer = self.manufacturer
        listNode = self._ref_node.find("serviceList", namespaces=self._namespaces)
        if listNode is not None:
            resultList = [ UpnpDevice1Service(manufacturer, child, namespaces=self._namespaces) for child in listNode ]
        return resultList

    @property
    def deviceType(self) -> str:
        """
            Returns the device type of the device.
        """
        rtnval = self._find_value("deviceType", namespaces=self._namespaces)
        return rtnval

    @property
    def friendlyName(self) -> str:
        """
            Returns the friendly name of the device.
        """
        rtnval = self._find_value("friendlyName", namespaces=self._namespaces)
        return rtnval

    @property
    def MACAddress(self) -> str:
        """
            Returns the MAC address of the device.
        """
        rtnval = self._find_value("MACAddress", namespaces=self._namespaces)
        return rtnval

    @property
    def manufacturer(self) -> str:
        """
            Returns the manufacturer name of the device.
        """
        rtnval = self._find_value("manufacturer", namespaces=self._namespaces)
        return rtnval

    @property
    def manufacturerURL(self) -> str:
        """
            Returns the manufacturer URL of the device.
        """
        rtnval = self._find_value("manufacturerURL", namespaces=self._namespaces)
        return rtnval

    @property
    def modelDescription(self) -> str:
        """
            Returns the model description of the device.
        """
        rtnval = self._find_value("modelDescription", namespaces=self._namespaces)
        return rtnval

    @property
    def modelName(self) -> str:
        """
            Returns the model name of the device.
        """
        rtnval = self._find_value("modelName", namespaces=self._namespaces)
        return rtnval

    @property
    def modelNumber(self) -> str:
        """
            Returns the model number of the device.
        """
        rtnval = self._find_value("modelNumber", namespaces=self._namespaces)
        return rtnval

    @property
    def modelURL(self) -> str:
        """
            Returns the model URL of the device.
        """
        rtnval = self._find_value("modelURL", namespaces=self._namespaces)
        return rtnval

    @property
    def serialNumber(self) -> str:
        """
            Returns the serial number of the device.
        """
        rtnval = self._find_value("serialNumber", namespaces=self._namespaces)
        return rtnval

    @property
    def UDN(self) -> str:
        """
            Returns the UDN of the device.
        """
        rtnval = self._find_value("UDN", namespaces=self._namespaces)
        return rtnval

    @property
    def UPC(self) -> str:
        """
            Returns the UPC of the device.
        """
        rtnval = self._find_value("UPC", namespaces=self._namespaces)
        return rtnval

    @property
    def presentationURL(self) -> str:
        """
            Returns the presentation URL of the device.
        """
        rtnval = self._find_value("presentationURL", namespaces=self._namespaces)
        return rtnval

    def to_dict(self, brief=False) -> dict:
        """
            Converts the icon data to a python dictionary.
        """
        dval = {}

        for nxt_child in self._ref_node:
            # We only use the generic value processing for
            # children that are singular
            if len(nxt_child) == 0:
                nxt_tag_full = nxt_child.tag
                start_index = nxt_tag_full.find("}") + 1
                nxt_tag = nxt_tag_full[start_index:]
                dval[nxt_tag] = nxt_child.text

        firstIcon = None
        if not brief:
            device_list = []
            for dvc in self.deviceList:
                lival = dvc.to_dict()
                device_list.append(lival)
            dval["deviceList"] = device_list

            icon_list = []
            for icon in self.iconList:
                lival = icon.to_dict()
                icon_list.append(lival)
                if firstIcon is None:
                    firstIcon = lival
            dval["iconList"] = icon_list

            service_list = []
            for svc in self.serviceList:
                lsval = svc.to_dict()
                service_list.append(lsval)
            dval["serviceList"] = service_list
        else:
            for icon in self.iconList:
                lival = icon.to_dict()
                if firstIcon is None:
                    firstIcon = lival
                    break

        dval["firstIcon"] = firstIcon

        return dval



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class RenderingControl1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'RenderingControl1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-upnp-org:service:RenderingControl:1'

    SERVICE_DEFAULT_VARIABLES = {
        "AudioDelay": { "data_type": "string", "default": None, "allowed_list": None},
        "AudioDelayLeftRear": { "data_type": "string", "default": None, "allowed_list": None},
        "AudioDelayRightRear": { "data_type": "string", "default": None, "allowed_list": None},
        "Bass": { "data_type": "i2", "default": None, "allowed_list": None},
        "DialogLevel": { "data_type": "string", "default": None, "allowed_list": None},
        "EQValue": { "data_type": "i2", "default": None, "allowed_list": None},
        "HeadphoneConnected": { "data_type": "boolean", "default": None, "allowed_list": None},
        "Loudness": { "data_type": "boolean", "default": None, "allowed_list": None},
        "MusicSurroundLevel": { "data_type": "string", "default": None, "allowed_list": None},
        "Mute": { "data_type": "boolean", "default": None, "allowed_list": None},
        "NightMode": { "data_type": "boolean", "default": None, "allowed_list": None},
        "OutputFixed": { "data_type": "boolean", "default": None, "allowed_list": None},
        "PresetNameList": { "data_type": "string", "default": None, "allowed_list": None},
        "RoomCalibrationAvailable": { "data_type": "boolean", "default": None, "allowed_list": None},
        "RoomCalibrationCalibrationMode": { "data_type": "string", "default": None, "allowed_list": None},
        "RoomCalibrationCoefficients": { "data_type": "string", "default": None, "allowed_list": None},
        "RoomCalibrationEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "RoomCalibrationID": { "data_type": "string", "default": None, "allowed_list": None},
        "SpeakerSize": { "data_type": "ui4", "default": None, "allowed_list": None},
        "SubCrossover": { "data_type": "string", "default": None, "allowed_list": None},
        "SubEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "SubGain": { "data_type": "string", "default": None, "allowed_list": None},
        "SubPolarity": { "data_type": "string", "default": None, "allowed_list": None},
        "SupportsOutputFixed": { "data_type": "boolean", "default": None, "allowed_list": None},
        "SurroundEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "SurroundLevel": { "data_type": "string", "default": None, "allowed_list": None},
        "SurroundMode": { "data_type": "string", "default": None, "allowed_list": None},
        "Treble": { "data_type": "i2", "default": None, "allowed_list": None},
        "Volume": { "data_type": "ui2", "default": None, "allowed_list": None},
        "VolumeDB": { "data_type": "i2", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "LastChange": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_GetBass(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetBass action.

            :returns: "CurrentBass"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetBass", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentBass",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetEQ(self, InstanceID, EQType, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetEQ action.

            :returns: "CurrentValue"
        """
        arguments = {
            "InstanceID": InstanceID,
            "EQType": EQType,
        }

        out_params = self.call_action("GetEQ", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetHeadphoneConnected(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetHeadphoneConnected action.

            :returns: "CurrentHeadphoneConnected"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetHeadphoneConnected", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentHeadphoneConnected",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetLoudness(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetLoudness action.

            :returns: "CurrentLoudness"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetLoudness", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentLoudness",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetMute(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetMute action.

            :returns: "CurrentMute"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetMute", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentMute",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetOutputFixed(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetOutputFixed action.

            :returns: "CurrentFixed"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetOutputFixed", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentFixed",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetRoomCalibrationStatus(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetRoomCalibrationStatus action.

            :returns: "RoomCalibrationEnabled", "RoomCalibrationAvailable"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetRoomCalibrationStatus", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RoomCalibrationEnabled", "RoomCalibrationAvailable",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetSupportsOutputFixed(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetSupportsOutputFixed action.

            :returns: "CurrentSupportsFixed"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetSupportsOutputFixed", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentSupportsFixed",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetTreble(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetTreble action.

            :returns: "CurrentTreble"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("GetTreble", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentTreble",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetVolume(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetVolume action.

            :returns: "CurrentVolume"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetVolume", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetVolumeDB(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetVolumeDB action.

            :returns: "CurrentVolume"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetVolumeDB", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("CurrentVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetVolumeDBRange(self, InstanceID, Channel, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetVolumeDBRange action.

            :returns: "MinValue", "MaxValue"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        out_params = self.call_action("GetVolumeDBRange", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("MinValue", "MaxValue",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_RampToVolume(self, InstanceID, Channel, RampType, DesiredVolume, ResetVolumeAfter, ProgramURI, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RampToVolume action.

            :returns: "RampTime"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "RampType": RampType,
            "DesiredVolume": DesiredVolume,
            "ResetVolumeAfter": ResetVolumeAfter,
            "ProgramURI": ProgramURI,
        }

        out_params = self.call_action("RampToVolume", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("RampTime",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ResetBasicEQ(self, InstanceID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ResetBasicEQ action.

            :returns: "Bass", "Treble", "Loudness", "LeftVolume", "RightVolume"
        """
        arguments = {
            "InstanceID": InstanceID,
        }

        out_params = self.call_action("ResetBasicEQ", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Bass", "Treble", "Loudness", "LeftVolume", "RightVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_ResetExtEQ(self, InstanceID, EQType, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the ResetExtEQ action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "EQType": EQType,
        }

        self.call_action("ResetExtEQ", arguments=arguments, aspects=aspects)

        return

    def action_RestoreVolumePriorToRamp(self, InstanceID, Channel, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the RestoreVolumePriorToRamp action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
        }

        self.call_action("RestoreVolumePriorToRamp", arguments=arguments, aspects=aspects)

        return

    def action_SetBass(self, InstanceID, DesiredBass, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetBass action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredBass": DesiredBass,
        }

        self.call_action("SetBass", arguments=arguments, aspects=aspects)

        return

    def action_SetChannelMap(self, InstanceID, ChannelMap, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetChannelMap action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "ChannelMap": ChannelMap,
        }

        self.call_action("SetChannelMap", arguments=arguments, aspects=aspects)

        return

    def action_SetEQ(self, InstanceID, EQType, DesiredValue, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetEQ action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "EQType": EQType,
            "DesiredValue": DesiredValue,
        }

        self.call_action("SetEQ", arguments=arguments, aspects=aspects)

        return

    def action_SetLoudness(self, InstanceID, Channel, DesiredLoudness, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetLoudness action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredLoudness": DesiredLoudness,
        }

        self.call_action("SetLoudness", arguments=arguments, aspects=aspects)

        return

    def action_SetMute(self, InstanceID, Channel, DesiredMute, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetMute action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredMute": DesiredMute,
        }

        self.call_action("SetMute", arguments=arguments, aspects=aspects)

        return

    def action_SetOutputFixed(self, InstanceID, DesiredFixed, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetOutputFixed action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredFixed": DesiredFixed,
        }

        self.call_action("SetOutputFixed", arguments=arguments, aspects=aspects)

        return

    def action_SetRelativeVolume(self, InstanceID, Channel, Adjustment, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRelativeVolume action.

            :returns: "NewVolume"
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "Adjustment": Adjustment,
        }

        out_params = self.call_action("SetRelativeVolume", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("NewVolume",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_SetRoomCalibrationStatus(self, InstanceID, RoomCalibrationEnabled, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRoomCalibrationStatus action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "RoomCalibrationEnabled": RoomCalibrationEnabled,
        }

        self.call_action("SetRoomCalibrationStatus", arguments=arguments, aspects=aspects)

        return

    def action_SetRoomCalibrationX(self, InstanceID, CalibrationID, Coefficients, CalibrationMode, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetRoomCalibrationX action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "CalibrationID": CalibrationID,
            "Coefficients": Coefficients,
            "CalibrationMode": CalibrationMode,
        }

        self.call_action("SetRoomCalibrationX", arguments=arguments, aspects=aspects)

        return

    def action_SetTreble(self, InstanceID, DesiredTreble, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetTreble action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "DesiredTreble": DesiredTreble,
        }

        self.call_action("SetTreble", arguments=arguments, aspects=aspects)

        return

    def action_SetVolume(self, InstanceID, Channel, DesiredVolume, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetVolume action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredVolume": DesiredVolume,
        }

        self.call_action("SetVolume", arguments=arguments, aspects=aspects)

        return

    def action_SetVolumeDB(self, InstanceID, Channel, DesiredVolume, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the SetVolumeDB action.
        """
        arguments = {
            "InstanceID": InstanceID,
            "Channel": Channel,
            "DesiredVolume": DesiredVolume,
        }

        self.call_action("SetVolumeDB", arguments=arguments, aspects=aspects)

        return



"""
.. module:: templates
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Package marking the folder that contains html and other renderable
               content templates.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

from akit.exceptions import AKitRuntimeError

DIR_TEMPLATES = os.path.dirname(os.path.abspath(__file__))

DIR_TAB_TEMPLATES = os.path.join(DIR_TEMPLATES, "tabs")

TABS_TEMPLATE_IMAGES = os.path.join(DIR_TAB_TEMPLATES, "tab-images.html")
TABS_TEMPLATE_SOUNDS = os.path.join(DIR_TAB_TEMPLATES, "tab-sounds.html")

class SummaryTemplateNames:
    Html = "html-template"
    Script = "script-template"
    Style = "style-template"




"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class CloudProxy1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:CloudProxy:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'CloudProxy1'

    SERVICE_DEFAULT_VARIABLES = {
        "DeviceList": { "data_type": "string", "default": None, "allowed_list": None},
        "ProxyList": { "data_type": "string", "default": None, "allowed_list": None},
        "UCSList": { "data_type": "string", "default": None, "allowed_list": None},
    }

    SERVICE_EVENT_VARIABLES = {
        "CloudProxyUpdate": { "data_type": "string", "default": None, "allowed_list": None},
    }

    def action_AddProxyDevice(self, DeviceId, UserAtCloud, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddProxyDevice action.

            :returns: "DeviceJID"
        """
        arguments = {
            "DeviceId": DeviceId,
            "UserAtCloud": UserAtCloud,
        }

        out_params = self.call_action("AddProxyDevice", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("DeviceJID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_AddUCSAccount(self, UserAtCloud, Port, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddUCSAccount action.

            :returns: "UCSJID", "Password"
        """
        arguments = {
            "UserAtCloud": UserAtCloud,
            "Port": Port,
        }

        out_params = self.call_action("AddUCSAccount", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("UCSJID", "Password",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DeleteProxyDevice(self, DeviceJID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteProxyDevice action.
        """
        arguments = {
            "DeviceJID": DeviceJID,
        }

        self.call_action("DeleteProxyDevice", arguments=arguments, aspects=aspects)

        return

    def action_DeleteUCSAccount(self, BareJID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeleteUCSAccount action.
        """
        arguments = {
            "BareJID": BareJID,
        }

        self.call_action("DeleteUCSAccount", arguments=arguments, aspects=aspects)

        return

    def action_GetDeviceList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetDeviceList action.

            :returns: "DeviceList"
        """
        arguments = { }

        out_params = self.call_action("GetDeviceList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("DeviceList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetProxyList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetProxyList action.

            :returns: "ProxyList"
        """
        arguments = { }

        out_params = self.call_action("GetProxyList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("ProxyList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetUCSList(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetUCSList action.

            :returns: "UCSList"
        """
        arguments = { }

        out_params = self.call_action("GetUCSList", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("UCSList",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import List, Optional

import csv
import netifaces
import os

ARPING_RANGE_CMD = """/bin/bash -c 'for i in {1..254} ;do ping %s.$i -w 5 -c 1 > /dev/null 2>&1 & echo "%s.$i" & done; wait < <(jobs -p); echo Done' """

def get_arp_table(normalize_hwaddr: bool=False):
    arp_table = {}

    with open('/proc/net/arp') as aif:
        #'IP address', 'HW type', 'Flags', 'HW address', 'Mask', 'Device'
        reader = csv.reader(aif, skipinitialspace=True, delimiter=' ')
        table_data_rows = [r for r in reader][1:]
        for ip, hwtype, flags, hwaddr, mask, ifname in table_data_rows:
            if normalize_hwaddr:
                hwaddr = hwaddr.replace(":", "").upper()
            iinfo = { hwaddr: {"hwaddr": hwaddr, "ip": ip, "ifname": ifname, "hwtype": hwtype, "flags": flags, "mask": mask }}
            arp_table.update(iinfo)

    return arp_table

def refresh_arp_table(exclude_interfaces: List=["lo"], include_interfaces: Optional[List[str]]=None):
    """
        ping -c 5 -b 10.x.x.255
    """

    interface_list = None
    if include_interfaces is not None:
        interface_list = include_interfaces
    else:
        interface_list = netifaces.interfaces()
    
    for ifname in interface_list:
        if ifname not in exclude_interfaces:
            address_info = netifaces.ifaddresses(ifname)
            if address_info is not None:

                # First look for IPv4 address information
                if netifaces.AF_INET in address_info:
                    addr_info = address_info[netifaces.AF_INET][0]
                    ifaddress = addr_info["addr"]
                    ifaddr_parts = ifaddress.split(".")
                    addr_prefix = ".".join(ifaddr_parts[:-1])
                    ping_cmd = ARPING_RANGE_CMD % (addr_prefix, addr_prefix)
                    os.system(ping_cmd)

    return

if __name__ == "__main__":
    refresh_arp_table()